#!/usr/bin/env python3
# coding=utf-8
"""
A script to import and link data to ODM.

See http://odm-user-guide.rtfd.io/en/latest/doc-odm-user-guide/import-data-using-python-script.html
for detailed guide
"""

from __future__ import division, print_function

import argparse
import collections
import copy
import csv
import itertools
import json
import pprint
import re
import sys
from time import sleep

try:
    # Python 3
    from urllib.parse import urlparse
except ImportError:
    # PY3TODO - this is a label for code that could be simplified/improved when
    # dropping support for Python 2
    # Python 2
    from urlparse import urlparse

import requests

# let's do it `six`-style! (https://github.com/benjaminp/six/blob/master/six.py#L35)
PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3

# PY3TODO
if PY3:
    bytes_type = bytes
else:
    bytes_type = str

ACC_RE = re.compile(r'([A-Z]+[0-9]+)$')

# regexp to match accession in square brackets appended to links
TRAILING_ACC_RE = r'\[([A-Z]+[0-9]+)\]$'
TRAILING_ACC_RE = re.compile(TRAILING_ACC_RE)

LOAD_RETRIES = 60
INTEGRATION_PREFIX = "integrationCurator/%s/integration/link/"
FILE_ENDPOINTS_DICT = {"study": "studyCurator/%s/studies",
                       "samples": "sampleCurator/%s/samples",
                       "variant": "variantCurator/%s/variant/vcf",
                       "expression": "expressionCurator/%s/expression/gct",
                       "flow-cytometry": "flowCytometryCurator/%s/flow-cytometry/facs"}
LINK_ENDPOINTS_DICT = {"samples_to_study": "sample/%s/to/study/%s",
                       "variant_to_sample": "variant/%s/to/sample/%s",
                       "flow-cytometry_to_sample": "flow-cytometry/%s/to/sample/%s",
                       "expression_to_sample": "expression/%s/to/sample/%s"}  # type: dict
LINK_ENDPOINTS_DICT = {k: INTEGRATION_PREFIX + v
                       for k, v in LINK_ENDPOINTS_DICT.items()}

# how long to wait for ETL jobs to be finished (in seconds)
ETL_WAITING_TIMEOUT = 30 * 60
# how often to poll for jobs
ETL_POLLING_INTERVAL = 5

# all supported file "sources" supported by ETL
ETL_SOURCES = ('S3', 'HTTP', 'Arvados', 'LOCAL')

# a mapping of URL scheme to its *default* ETL "source" (see above)
# if no source is provided explicitly
SCHEME_TO_ETL_SOURCE = {
    'http': 'HTTP',
    'https': 'HTTP',
    's3': 'S3',
    'file': 'LOCAL'
}

# guard against possible mistakes in ETL source constants
assert set(SCHEME_TO_ETL_SOURCE.values()).issubset(ETL_SOURCES)

COMMON_URL_PREFIX = 'frontend/rs/genestack'

TAGS = {"-sm": "samples",
        "-lb": "libraries",
        "-pr": "preparations",
        "-f": "flow-cytometry",
        "-fm": "flow-cytometry-metadata",
        "-v": "variant",
        "-vm": "variant-metadata",
        "-e": "expression",
        "-em": "expression-metadata",
        "-mpf": "mapping-file",
        "-mpfa": "mapping-file-accession",
        "-mpfm": "mapping-file-metadata",
        "-nfa": "number-of-feature-attributes",
        "-dc": "data-class",
        "-ms": "measurement-separator"
        }

LIB_PREP_TAGS = {'libraries', 'preparations'}
SIGNAL_TAGS = {'flow-cytometry', 'variant', 'expression'}

FILE_FOUND_ERR_MSG_RE = r'job instance already exists .+ jobExecId=([0-9]+)'
FILE_FOUND_ERR_MSG_RE = re.compile(FILE_FOUND_ERR_MSG_RE)


def is_acc(string):
    return bool(ACC_RE.match(string))


def parse_file_signal(link):
    """ Parse URLs like http://examp.le/file.txt[GSF123456] """
    version = TRAILING_ACC_RE.search(link)
    if version is not None:
        link = link[:version.start()]
        version = version.groups()[0]
    return link, version


def partition(pred, iterable):
    "Use a predicate to partition entries into false entries and true entries"
    # partition(is_odd, range(10)) --> 0 2 4 6 8   and  1 3 5 7 9
    t1, t2 = itertools.tee(iterable)
    return itertools.filterfalse(pred, t1), filter(pred, t2)


class GroupLinkingError(Exception):
    """
    Exception raised for error during group linking. E.g linking signals to
    samples or samples to study
    """

    def __init__(self, what, accession_to, accession_from):
        self.what = what
        self.accession_to = accession_to
        self.accession_from = accession_from
        # PY3TODO
        super(GroupLinkingError, self).__init__(self, what, accession_to, accession_from)


def green_text(text):
    if sys.platform == 'win32' or not sys.stdout.isatty():
        return text
    # PY3TODO
    return u"""\u001b[32m""" + text + u"""\u001b[0m"""


def red_text(text):
    if sys.platform == 'win32' or not sys.stdout.isatty():
        return text
    # PY3TODO
    return u"""\u001b[31m""" + text + u"""\u001b[0m"""


class TemplateAccessionSupplier:
    """
    Container for TEMPLATE_ACCESSION.
    If empty, could try to request default template's accession
    """
    def __init__(self, args):
        self.server = args.SERVER
        self.token = args.API_TOKEN
        self.debug = args.debug
        self.template_accession = args.TEMPLATE_ACCESSION
        self.should_request_default = args.TEMPLATE_ACCESSION is None and \
                                      args.study_accession is None
        self._validate()

    def _validate(self):
        if self.token is None and self.should_request_default:
            _err("Please provide template accession when using Access Token", in_red=True)
            sys.exit(1)

    def __call__(self, *args, **kwargs):
        if self.should_request_default:
            self.template_accession = self._request_default_template()
            self.should_request_default = False
        return self.template_accession

    def _authenticate(self):
        aut_url = "/frontend/endpoint/application/invoke/genestack/signin/authenticateByApiToken"
        url_authenticate = self.server + aut_url
        session = requests.Session()
        try:
            result = session.post(url_authenticate, json=[self.token], timeout=10)
            if not json.loads(result.text)["result"]["authenticated"]:
                if self.debug:
                    print(result, result.text)
                print(red_text("You set invalid token or your license is expired!"))
                sys.exit(1)
            return session
        except requests.exceptions.ConnectionError as error:
            if not self.debug:
                if error.__class__.__name__ == 'ConnectionError':
                    _err("Failed connection to server {srv}. Check the server name"
                         .format(srv=self.server))
                if error.__class__.__name__ == 'ConnectTimeout':
                    _err("Connection timeout to server {srv}".format(srv=self.server))
            else:
                _err("{err_name}: {err_args}"
                     "".format(err_name=error.__class__.__name__,
                               err_args=error.args))
            sys.exit(1)

    def _request_default_template(self):
        url_request_template = (self.server +
                                "/frontend/endpoint/application/invoke/genestack"
                                "/study-metainfotemplateeditor/listTemplates")
        session = self._authenticate()
        response = session.post(url=url_request_template, json=[])
        assert response.status_code == 200, response.text
        assert "result" in response.text, response.text
        result = json.loads(response.text)["result"]
        accession = [template["accession"] for template in result
                     if template.get("isDefault", False)][0]
        assert accession
        return accession


def get_study(params):
    response = requests.get(
        url='{0}/frontend/rs/genestack/studyUser/{1}/studies/{2}'.format(
            params.SERVER, params.APP_VERSION, params.study_accession),
        headers=params.headers)
    return response


def check_access_to_study(params):
    response = get_study(params)
    if response.status_code != 200:
        if params.debug:
            print(response, response.text)
        _err("You don't have access permissions to study: "
             "%s or it doesn't exist!" % params.study_accession, in_red=True)
        sys.exit(1)


def read_utf_csv(csv_data):
    ''' Yield stripped and decoded CSV rows as dictionaries '''
    # PY3TODO
    if sys.version_info.major == 3:
        csv_data = (line.decode('utf-8') for line in csv_data)
    for row in csv.DictReader(csv_data):
        yield {k.strip(): v.strip() for k, v in row.items()}


def fetch_and_parse_csv(csv_link):
    """ Fetch remote one-row CSV file and parse it to a dictionary """
    # must not be empty or None
    assert csv_link
    try:
        response = requests.get(csv_link)
    except requests.exceptions.RequestException as error:
        _err("Error accessing metadata file '{link}':\n{err_name}: {err_args}"
             "".format(link=csv_link, err_name=error.__class__.__name__,
                       err_args=error.args))
        sys.exit(1)
    if not response.ok:
        _err("Error accessing metadata file '{}'".format(csv_link),
             response=response, in_red=True)
        sys.exit(1)
    csv_data = read_utf_csv(response.iter_lines())
    try:
        metadata = next(csv_data)
    # pylint: disable=broad-except
    except Exception as err:
        _err("Error parsing '{}' as CSV file:\n"
             "{}: {}".format(csv_link, err, err.args),
             in_red=True)
        sys.exit(1)
    try:
        next(csv_data)
        _err("Metadata file '{}' contains multiple rows, using only the first one"
             "".format(csv_link), in_red=True)
    except StopIteration:
        pass
    return metadata


def _get_etl_source_from_url(url):
    parsed_url = urlparse(url)
    source = SCHEME_TO_ETL_SOURCE.get(parsed_url.scheme)
    if source is None:
        _err("Scheme '{}' is not supported, exiting".format(parsed_url.scheme),
             in_red=True)
        sys.exit(1)
    return source


def link_mappings(mapping_file_acc, expression_acc, params):
    """ Link gene-tx mapping to expression data """
    url = '{}/{}/integrationCurator/{}/links'.format(params.SERVER, COMMON_URL_PREFIX,
                                                     params.APP_VERSION)
    payload = [{'firstId': expression_acc,
                'firstType': 'expressionGroup',
                'secondId': mapping_file_acc,
                'secondType': 'geneTranscriptMapping'}]
    resp = requests.post(url, headers=params.headers, json=payload)
    if not resp.ok:
        _err("Linking gene-transcription mapping '{}' to expression group '{}' failed"
             "".format(mapping_file_acc, expression_acc),
             response=resp, in_red=True)
        sys.exit(1)
    print("Successfully linked: [mapping_file_to_expression]")


def check_mapping_file(acc, params):
    """
    Check if provided accession corresponds to an existing mapping file

    Returns the same accession if it's fine, otherwise exits with error
    """
    if acc is None:
        return None
    url = ('{}/{}/reference-data/{}/xrefsets/{}/metadata'
           ''.format(params.SERVER, COMMON_URL_PREFIX, params.APP_VERSION, acc))
    resp = requests.get(url, headers=params.headers)
    if resp.ok:
        return acc
    _err("mapping file with the accession '{}' is not found".format(acc),
         in_red=True)
    sys.exit(1)


def _err(msg, response=None, in_red=False):
    ''' Print error message '''
    # PY3TODO
    if isinstance(msg, bytes_type):
        msg = msg.decode('utf-8', errors='replace')
    if in_red:
        msg = red_text(msg)
    if response is not None:
        text = response.text
        req = response.request
        if isinstance(text, bytes_type):
            text = text.decode('utf-8', errors='ignore')
        try:
            text = pprint.pformat(json.loads(text))
        except ValueError:
            pass
        msg += (u"\nStatus code: {}\n"
                "Request URL: {}\n"
                "Request body: '{}'\n"
                "Error message from server:\n{}".format(response.status_code,
                                                        req.url,
                                                        req.body,
                                                        text))
    print(msg, file=sys.stderr)


def _check_job_error(response):
    ''' Checks if import job error is actually a message about duplicate '''
    # if a file had already been submitted, REST API returns 409 Conflict
    # pylint: disable=no-member
    if response.status_code != requests.codes.CONFLICT:
        return None
    r_data = _unpack_job_response(response)
    job_id_match = FILE_FOUND_ERR_MSG_RE.search(r_data['error']['message'])
    if job_id_match is None:
        return None
    job_id = int(job_id_match.groups()[0])
    return job_id


def _get_mdata_params(metadata_link, inline_metadata):
    ''' Return dictionary of metadata parameters

    If ``inline_metadata`` is True, parse provided CSV link and return its content
    as a "subdictionary".
    Otherwise, pass metadata link to ETL as-is
    '''
    if not inline_metadata:
        return {"metadataLink": metadata_link}
    metadata = fetch_and_parse_csv(metadata_link)
    return {"data": metadata}


def _select_source(links):
    sources = set(_get_etl_source_from_url(link) for link in links)
    if len(sources) != 1:
        _err("Mixing URL schemes is not supported; provide source explicitly",
             in_red=True)
        sys.exit(1)
    assert sources != set([None])
    return sources.pop()


# pylint: disable-next=too-many-arguments
def _prepare_etl_payload(kind, metadata_link, template_id=None, data_link=None,
                         prev_version=None, number_of_feature_attributes=None,
                         data_class=None, measurement_separator=None, source=None):
    ''' Prepare payload to be sent to ETL as parameters '''
    payload = {}
    if template_id is not None:
        payload["templateId"] = template_id
    inline_metadata = False
    if metadata_link is not None:
        inline_metadata = (kind == 'transcript-mapping')
        md_params = _get_mdata_params(metadata_link, inline_metadata)
        payload.update(md_params)
    if data_link is not None:
        payload["dataLink"] = data_link
    if prev_version is not None:
        payload["previousVersion"] = prev_version
    if number_of_feature_attributes is not None:
        payload["numberOfFeatureAttributes"] = number_of_feature_attributes
    if data_class is not None:
        payload["dataClass"] = data_class
    if measurement_separator is not None:
        payload["measurementSeparator"] = measurement_separator.lstrip()
    if source is None:
        # determine list of links eligible for source guesswork
        source_links = [data_link]
        if not inline_metadata:
            source_links.append(metadata_link)
        source = _select_source(filter(lambda l: l is not None, source_links))
    payload["source"] = source
    return payload


# pylint: disable-next=too-many-arguments
def _async_import(kind, metadata_link, params, data_link=None,
                  prev_version=None, number_of_feature_attributes=None,
                  data_class=None, measurement_separator=None):
    ''' Import data using Job/ETL API

    Returns ``job_info`` dictionary with all the submitted (and finished) job
    information, and a ``file_exists`` boolean: ``True`` if file has been found
    in ODM, and ``False`` if it has been uploaded anew.
    '''
    if metadata_link is None and kind in ('study', 'samples'):
        _err("No metadata link is provided for {} import"
             "".format(kind))
        sys.exit(1)
    if data_link is None and kind in ('expression', 'variant', 'flow-cytometry',
                                      'transcript-mapping'):
        _err("No data link is provided for {} import"
             "".format(kind))
        sys.exit(1)
    url = "{}/{}/job/{}/import/{}/".format(
        params.SERVER, COMMON_URL_PREFIX, params.APP_VERSION, kind.replace('_', '-'))
    if params.ALLOW_DUPLICATES:
        url += '?allow_dups=true'
    template_id = params.TEMPLATE_ACCESSION_SUPPLIER()
    payload = _prepare_etl_payload(kind, metadata_link, template_id, data_link,
                                   prev_version, number_of_feature_attributes,
                                   data_class, measurement_separator, params.ETL_SOURCE)

    resp = requests.post(url, headers=params.headers, json=payload)
    if resp.status_code == 200:
        r_data = _unpack_job_response(resp)
        job_id = int(r_data.get('jobExecId'))
        file_exists = False
    else:
        job_id = _check_job_error(resp)
        if job_id is None:
            _err("Submitting job failed!", response=resp, in_red=True)
            sys.exit(1)
        file_exists = True
    job_info = _get_finished_job_info(job_id, params)
    return job_info, file_exists


def _unpack_job_response(r):
    # TODO: go through other unguarded `json.loads` calls and wrap them too
    try:
        return json.loads(r.text)
    except json.decoder.JSONDecodeError:
        _err("Unexpected response received from server "
             "(cannot decode body as JSON)", response=r, in_red=True)
        sys.exit(1)


def _get_finished_job_info(job_id, params):
    ''' Wait for job to finish and return result '''
    job_url = "{}/{}/job/{}/{}".format(
        params.SERVER, COMMON_URL_PREFIX, params.APP_VERSION, job_id)
    info_url = job_url + "/info"
    output_url = job_url + "/output"
    total_tries = params.JOB_TIMEOUT // ETL_POLLING_INTERVAL
    progress_reported = False
    for attempt in range(total_tries):
        status = requests.get(info_url, headers=params.headers)
        status = _unpack_job_response(status).get(u'status')
        if status not in (u'STARTED', u'STARTING', u'RUNNING'):
            break
        sleep(ETL_POLLING_INTERVAL)
        # report "progress" after two minutes of waiting
        if attempt * ETL_POLLING_INTERVAL > 120:
            progress_reported = True
            print("Waiting for job #{} to complete, {} minutes before timeout"
                  "".format(job_id, (total_tries - attempt) * ETL_POLLING_INTERVAL // 60 + 1),
                  end='\r', file=sys.stderr)
    else:
        # clear screen after redrawn messages
        _err('')
        _err("Job {} has been running for {} seconds with no result, exiting"
             "".format(job_id, params.JOB_TIMEOUT))
        sys.exit(1)
    # clear screen after redrawn messages if they had been printed
    if progress_reported:
        _err('')
    output_r = requests.get(output_url, headers=params.headers)
    output = _unpack_job_response(output_r)
    if status != u'COMPLETED':
        _err("Job {} failed with status {}".format(job_id, status),
             response=output_r, in_red=True)
        sys.exit(1)
    return output


def add_mappings(data_link, params, metadata_link=None):
    """ Upload gene-tx mapping file with optional metadata
        for import xref-mappings we are using syncronous endpoint
        https://genestack.atlassian.net/browse/ODM-7489
    """
    metadata = fetch_and_parse_csv(metadata_link) if metadata_link else None
    url = ('{}/{}/reference-data/{}/xrefsets/'
           ''.format(params.SERVER, COMMON_URL_PREFIX, params.APP_VERSION))
    source = _get_etl_source_from_url(data_link).lower()
    payload = {
        "dataLink": data_link,
        "dataSource": source,
        "xrefSetType": "gene-transcript"
    }

    if metadata is not None:
        payload["metadata"] = metadata

    resp = requests.post(url, headers=params.headers, json=payload)
    if resp.status_code != 200:
        _err("Importing '{}' failed!".format(data_link),
             response=resp, in_red=True)
        sys.exit(1)
    accession = json.loads(resp.text)['xrefSetId']
    if accession is None:
        _err("Unexpected response from the server while importing '{}'."
             .format(data_link), response=resp, in_red=True)
        sys.exit(1)
    print("Mapping file {} was added successfully".format(accession))
    return accession


def add_study(params):
    ''' Create new study or return existing one '''
    if params.study_accession:
        check_access_to_study(params)
        return params.study_accession
    job_info, study_exists = _async_import(
        kind='study',
        metadata_link=params.study_link,
        params=params
    )
    accession = job_info.get(u'result', {}).get(u'accession')
    if accession is None:
        _err("Mandatory file: study metadata wasn't loaded. Uploading is stopped!")
        sys.exit(1)
    if study_exists:
        # this error message is parsed by load.py script: do not change it
        _err("'{}' has already been uploaded as study '{}'"
             "".format(params.study_link, accession))
        if params.FAIL_IF_FILE_EXISTS:
            sys.exit(1)
    else:
        print("study " + accession + " was added successfully")
    return accession


def add_and_link_samples(params, study, sample_link):
    job_info, samples_exist = _async_import('samples', metadata_link=sample_link,
                                            params=params)
    group_acc = job_info.get(u'result', {}).get(u'groupAccession')
    if group_acc is None:
        _err("Mandatory file: samples metadata wasn't loaded. "
             "Without samples metadata file, linking of signals to samples are impossible. "
             "Uploading of signals files are stopped!", in_red=True)
        sys.exit(1)
    if samples_exist:
        _err("'{}' has already been uploaded as sample group '{}'"
             "".format(sample_link, group_acc))
        if params.FAIL_IF_FILE_EXISTS:
            sys.exit(1)
    else:
        print("samples were added successfully (sample group accession is {})"
              "".format(group_acc))
    link_by_parent(what="samples_to_study",
                   accession_from=group_acc,
                   accession_to=study,
                   params=params)
    return group_acc


def get_libs_preps_by_study(file_type, params, study):
    link_word = {'libraries': 'library', 'preparations': 'preparation'}[file_type]
    url = ('{}/{}/integrationCurator/{}/integration/link/{}/group/by/study/{}'
           ''.format(params.SERVER, COMMON_URL_PREFIX, params.APP_VERSION, link_word, study))
    response = requests.get(url, headers=params.headers)
    if response.ok:
        return json.loads(response.text)
    if params.debug:
        print(response, response.text)
    _err("You don't have access permissions to study: {} or it doesn't exist!"
         "".format(study), in_red=True)
    sys.exit(1)


def check_study_has_libs_preps(group_acc, samples_group, file_type, params, study):
    libs_preps = get_libs_preps_by_study(file_type, params, study)
    # example of libs_preps: [{"itemId":"GSF020175","metadata":{}}]
    accessions = {item['itemId'] for item in libs_preps}
    if group_acc not in accessions:
        raise GroupLinkingError(what=file_type,
                                accession_to=group_acc,
                                accession_from=samples_group)


def add_and_link_libs_preps(
        samples_group,
        metadata_link,
        file_type,
        params,
        study,
        failures
):
    job_info, exists = _async_import(file_type, metadata_link=metadata_link,
                                     params=params)
    group_acc = job_info.get(u'result', {}).get(u'groupAccession')
    if exists:
        _err("'{}' has already been uploaded as {} group '{}'"
             "".format(metadata_link, file_type, group_acc))
        if params.FAIL_IF_FILE_EXISTS:
            sys.exit(1)
    else:
        print("{ft} were added successfully ({ft} group accession is {acc})"
              "".format(ft=file_type, acc=group_acc))
    link_by_parent(what="{}_to_samples".format(file_type),
                   accession_from=group_acc,
                   accession_to=samples_group,
                   params=params)
    try:
        # TODO remove this check when https://genestack.atlassian.net/browse/ODM-7793 will be fixed
        check_study_has_libs_preps(group_acc, samples_group, file_type, params, study)
    except GroupLinkingError as ex:
        _err("Linking {} ({} to {}) failed"
             "".format(file_type, group_acc, samples_group), in_red=True)
        if not params.IGNORE_LINKING_ERRORS:
            sys.exit(1)
        failures.append(ex)
        return group_acc
    print("Successfully linked: [{}_to_samples]".format(file_type))
    return group_acc


def add_signals(
        file_signal,
        file_metadata,
        file_type,
        number_of_feature_attributes,
        data_class,
        measurement_separator,
        params):
    data_link, prev_version = parse_file_signal(file_signal)
    job_info, signals_exist = _async_import(file_type, metadata_link=file_metadata,
                                            data_link=data_link,
                                            prev_version=prev_version,
                                            number_of_feature_attributes=number_of_feature_attributes,
                                            data_class=data_class,
                                            measurement_separator=measurement_separator,
                                            params=params)
    group_acc = job_info.get(u'result', {}).get(u'groupAccession')
    if signals_exist:
        signal_str = "'{}'".format(file_signal)
        if file_metadata is not None:
            signal_str = "'{} + '{}'".format(signal_str, file_metadata)
        _err("{} {} have already been uploaded as signal group '{}'"
             "".format(file_type, signal_str, group_acc))
        if params.FAIL_IF_FILE_EXISTS:
            sys.exit(1)
    else:
        print("{} data were added successfully (group accession is {})"
              "".format(file_type.replace('-', ' '), group_acc))
    return group_acc


def link_by_parent(what, accession_to, accession_from, params):
    ''' Link data using new group-centric API '''
    ENDPOINT_DICT = {
        'samples_to_study': 'sample/group/{sourceId}/to/study/{targetId}',
        'libraries_to_samples': 'library/group/{sourceId}/to/sample/group/{targetId}',
        'preparations_to_samples': 'preparation/group/{sourceId}/to/sample/group/{targetId}',
        'expression_to_sample': 'expression/group/{sourceId}/to/sample/group/{targetId}',
        'expression_to_libraries': 'expression/group/{sourceId}/to/library/group/{targetId}',
        'expression_to_preparations': 'expression/group/{sourceId}/to/preparation/group/{targetId}',
        'variant_to_sample': 'variant/group/{sourceId}/to/sample/group/{targetId}',
        # API not yet implemented
        # 'variant_to_libraries': 'variant/group/{sourceId}/to/library/group/{targetId}',
        # 'variant_to_preparations': 'variant/group/{sourceId}/to/preparation/group/{targetId}',
        'flow-cytometry_to_sample': 'flow-cytometry/group/{sourceId}/to/sample/group/{targetId}',
        # API not yet implemented
        # 'flow-cytometry_to_libraries':
        #     'flow_cytometry/group/{sourceId}/to/library/group/{targetId}',
        # 'flow-cytometry_to_preparations':
        #     'flow_cytometry/group/{sourceId}/to/preparation/group/{targetId}',
    }
    url = '/'.join([params.SERVER, COMMON_URL_PREFIX,
                    INTEGRATION_PREFIX.rstrip('/') % params.APP_VERSION, ENDPOINT_DICT[what]])
    response = requests.post(url.format(sourceId=accession_from,
                                        targetId=accession_to),
                             headers=params.headers)
    if response.ok:
        print("Successfully linked: [{}]".format(what))
        return

    what = what.replace('_', ' ')
    _err("Linking {what} ({accession_from} to {accession_to}) failed"
         "".format(**locals()), response=response, in_red=True)
    raise GroupLinkingError(what=what,
                            accession_to=accession_to,
                            accession_from=accession_from)


def check_mapping_files_arguments(signal_args):
    # check if there are no mappings and we can skip the checks
    if not any(s.startswith("mapping-file") for s in signal_args):
        return
    if "expression" not in signal_args:
        _err("mapping file is supported with expression matrices only",
             in_red=True)
        sys.exit(1)
    mpf_prefix = 'mapping-'
    mapping_args = collections.Counter(s[len(mpf_prefix):]
                                       for s in signal_args
                                       if s.startswith(mpf_prefix))
    total_mpfiles = mapping_args['file'] + mapping_args['file-accession']
    if total_mpfiles == 0:
        # we checked that there's *something* starting with `mpf_prefix`,
        # and that leaves only metadata
        _err("the parameter '-mpf' for mapping file loading is missing",
             in_red=True)
        sys.exit(1)
    if total_mpfiles > 1:
        _err("Only one mapping file is expected, "
             "check the value of parameters 'mpf' or 'mpfa'",
             in_red=True)
        sys.exit(1)
    if mapping_args["file-metadata"] and mapping_args["file-accession"]:
        _err("Cannot attach mapping metainfo to an existing mapping file",
             in_red=True)
        sys.exit(1)
    if mapping_args["file_metadata"] > 1:
        _err("Cannot attach multiple metadata sources to a mapping file",
             in_red=True)


class SaneArgumentParser(argparse.ArgumentParser):
    """Disables prefix matching in ArgumentParser."""

    # PY3TODO: `argparse.ArgumentParser(allow_abbrev=False)` introduced in 3.5

    def _get_option_tuples(self, option_string):
        """Prevent argument parsing from looking for prefix matches."""
        return []

    # pylint: disable-next=redefined-outer-name
    def parse_args(self, args=None, namespace=None):
        if args is None:
            # args default to the system args
            args = []
            prev = ""
            for a in sys.argv[1:]:
                # to bypass a bug in argparse library, add space to the front of the argument
                if (prev == "-ms" or prev == "--measurement-separator") and a.startswith('-'):
                    args.append(' ' + a)
                elif a.startswith('--'):
                    args.append(a.replace('_', '-'))
                else:
                    args.append(a)
                prev = a
        return super(SaneArgumentParser, self).parse_args(args, namespace)


# pylint: disable-next=protected-access
class _SingletonAction(argparse._StoreAction):
    """ Ensures that only one argument is passed """

    def __init__(self, mutually_exclusive, err_msg, *args, **kwargs):
        super(_SingletonAction, self).__init__(*args, **kwargs)
        self.err_msg = err_msg
        self.mutually_exclusive = mutually_exclusive

    def __call__(self, parser_, namespace, values, option_string=None):
        already_defined = any(getattr(namespace, v) is not None
                              for v in self.mutually_exclusive)
        if already_defined:
            raise argparse.ArgumentError(self, self.err_msg)
        setattr(namespace, self.dest, values)

# pylint: disable-next=protected-access
class _StoreServerName(argparse._StoreAction):
    """ Ensures that server name is not ended wit `/` """

    def __call__(self, parser_, namespace, values, option_string=None):
        setattr(namespace, self.dest, values.rstrip("/"))


def prevent_redundant_parameters(mutually_exclusive, err_msg):
    return lambda *args, **kwargs: _SingletonAction(mutually_exclusive, err_msg, *args, **kwargs)


class Headers(dict):
    """Derives http-headers from arguments"""

    def __init__(self, args):
        super(dict, self).__init__()
        self["Genestack-API-Token"] = args.API_TOKEN
        self["Authorization"] = "Bearer {}".format(args.ACCESS_TOKEN) if args.ACCESS_TOKEN else None
        self["Accept"] = "application/json"
        self["Content-Type"] = "application/json"
        self._validate()

    def _validate(self):
        if self["Genestack-API-Token"] and self["Authorization"]:
            _err("Please provide a Genestack-Api-Token or an Access Token but not both",
                 in_red=True)
            sys.exit(1)
        if self["Genestack-API-Token"] is None and self["Authorization"] is None:
            _err("Please provide a Genestack-Api-Token or an Access Token",
                 in_red=True)
            sys.exit(1)

    def __getstate__(self):
        return self


class DeprecatedAction(argparse.Action):
    """ Issues deprecation warning when called """

    def __call__(self, parser_, namespace, values, option_string=None):
        _err("'{}' is no longer supported, and left a no-op for "
             "compatibility reasons".format(option_string))


# PY3TODO
class ParserAstState(object):
    def __init__(self):
        self.sample_node_list = []
        self.current_node = None

    def has_libraries_or_preparations(self):
        all_tags = self.get_all_tags()
        return not all_tags.isdisjoint(LIB_PREP_TAGS)

    def get_all_tags(self):
        result = set()

        def get_tags(node):
            result.add(node['tag'])
            for child_node in node.get('children', []):
                get_tags(child_node)

        for node in self.sample_node_list:
            get_tags(node)
        return result

    def has_consistent_data_model(self):
        if not self.has_libraries_or_preparations():
            return True
        # in case libraries or preparations have been passed, all signals
        # should be linked to them instead of samples
        for sample_node in self.sample_node_list:
            for child_node in sample_node.get('children', []):
                if child_node['tag'] not in LIB_PREP_TAGS:
                    return False
        return True

    def has_non_expression_signals(self):
        all_tags = self.get_all_tags()
        return not all_tags.issubset({'samples', 'libraries', 'preparations',
                                      'expression', 'mapping-file'})


class BaseCustomAction(argparse.Action):
    def handle_action(self, tag, value, option_string):
        raise NotImplementedError('This method expected to be overrided in subclasses')

    def __call__(self, _parser, _args, value, option_string=None):
        opt_str = option_string.replace('--', '')
        tag = TAGS[opt_str] if opt_str.startswith('-') else opt_str
        if value is None:
            _err("You didn't provide {} file URL! Exit!"
                 "".format(tag.replace('-', ' ')), in_red=True)
            sys.exit(1)
        self.handle_action(tag, value, option_string)
        data = getattr(_args, self.dest)
        data = [] if data is None else data
        data.append(value)
        setattr(_args, self.dest, data)


def make_samples_action(parser_state):
    class SampleAction(BaseCustomAction):
        def handle_action(self, tag, value, option_string):
            new_sample_node = {'tag': tag, 'value': value}
            parser_state.sample_node_list.append(new_sample_node)
            parser_state.current_node = new_sample_node

    return SampleAction


def make_libraries_and_preparations_action(parser_state):
    class LibPrepAction(BaseCustomAction):
        def handle_action(self, tag, value, option_string):
            if is_acc(value):
                # (libraries|preparations) <accession>
                new_implicit_sample_node = {}
                new_implicit_sample_node['tag'] = 'samples'
                new_implicit_sample_node['value'] = 'implicit'
                current_sample_node = new_implicit_sample_node
                parser_state.sample_node_list.append(new_implicit_sample_node)
            else:
                # Search for the closest 'explicit' samples node to link the current lib/prep to it.
                current_sample_node = next(
                    (
                        node for node in reversed(parser_state.sample_node_list)
                        if node['value'] != 'implicit'
                    ),
                    None
                )
                if current_sample_node is None:
                    _err("You've provided {} before sample file or sample parent accession. Exit!"
                         .format(tag.replace('-', ' ')), in_red=True)
                    sys.exit(1)

            new_node = {'tag': tag, 'value': value}
            children = current_sample_node.get('children', [])
            children.append(new_node)
            current_sample_node['children'] = children
            parser_state.current_node = new_node

    return LibPrepAction


def make_signal_action(parser_state):
    class SignalAction(BaseCustomAction):
        def handle_action(self, tag, value, option_string):
            current_node = parser_state.current_node
            if current_node is None:
                _err("You've provided {} before sample file or sample parent accession. Exit!"
                     .format(tag.replace('-', ' ')), in_red=True)
                sys.exit(1)

            if tag.endswith('-metadata'):
                file_type = tag[:-9]
                children = current_node.get('children', [])
                # search for the closest node with the same file_type which
                # doesn't contain a metadata link
                last_node = next(
                    (
                        item for item in reversed(children)
                        if item['tag'] == file_type and 'metadata' not in item
                    ),
                    None
                )
                if last_node is None:
                    _err("No data link is provided for {} metadata '{}'"
                         "".format(file_type, value), in_red=True)
                    sys.exit(1)

                last_node['metadata'] = value
            elif tag == 'number-of-feature-attributes':
                children = current_node.get('children', [])
                last_node = next(
                    (
                        item for item in reversed(children)
                        if 'nfa' not in item
                    ),
                    None
                )
                if last_node is None:
                    _err(
                        "A file can only have one value of number of feature attributes. "
                        "Please check --number-of-feature-attributes (-nfa) argument in your input.",
                        in_red=True)
                    sys.exit(1)

                last_node['nfa'] = value
            elif tag == 'data-class':
                children = current_node.get('children', [])
                last_node = next(
                    (
                        item for item in reversed(children)
                        if 'dc' not in item
                    ),
                    None
                )
                if last_node is None:
                    _err(
                        "A file can only have one value of data class. "
                        "Please check --data-class (-dc) argument in your input.",
                        in_red=True)
                    sys.exit(1)

                last_node['dc'] = value
            elif tag == 'measurement-separator':
                children = current_node.get('children', [])
                last_node = next(
                    (
                        item for item in reversed(children)
                        if 'ms' not in item
                    ),
                    None
                )
                if last_node is None:
                    _err(
                        "A file can have only one value for the measurement separator. "
                        "Please check --measurement-separator (-ms) argument in your input.",
                        in_red=True)
                    sys.exit(1)

                last_node['ms'] = value
            else:
                new_signal_node = {'tag': tag, 'value': value}
                children = current_node.get('children', [])
                children.append(new_signal_node)
                current_node['children'] = children

    return SignalAction


def make_mapping_file_action(parser_state):
    class MappingFileAction(BaseCustomAction):
        def handle_action(self, tag, value, option_string):
            current_node = parser_state.current_node
            if current_node is None:
                _err("You provide " + tag.replace('-', ' ') +
                     " before sample file. Exit!", in_red=True)
                sys.exit(1)

            children = current_node.get('children', [])
            mapping_node = {'tag': tag, 'value': value}
            children.append(mapping_node)
            current_node['children'] = children

    return MappingFileAction


def check_and_merge_mapping_file_nodes(signal_nodes):
    non_mf_nodes_f, mf_nodes_f = partition(
        lambda node: node['tag'].startswith('mapping-file'),
        signal_nodes)
    mf_nodes = list(mf_nodes_f)
    if len(mf_nodes) == 0:
        return signal_nodes

    signal_args = []
    for node in signal_nodes:
        signal_args.append(node['tag'])
        signal_args.append(node['value'])
    check_mapping_files_arguments(signal_args)

    new_mf_node = {'tag': 'mapping-file'}
    for mf_node in mf_nodes:
        if mf_node['tag'] == 'mapping-file-metadata':
            new_mf_node['metadata'] = mf_node['value']
        else:
            new_mf_node['value'] = mf_node['value']
    return list(non_mf_nodes_f) + [new_mf_node]


def check_and_merge_mapping_file_in_sample_nodes(sample_nodes):
    for sample_node in sample_nodes:
        children = sample_node.get('children', [])
        children_tags = {child['tag'] for child in children}
        if children_tags.issubset(LIB_PREP_TAGS):
            for lib_prep_node in children:
                if 'children' not in lib_prep_node:
                    continue
                signal_nodes = lib_prep_node.get('children')
                lib_prep_node['children'] = check_and_merge_mapping_file_nodes(signal_nodes)
        else:
            if 'children' in sample_node:
                sample_node['children'] = check_and_merge_mapping_file_nodes(children)


def add_signals_to_parent(parent_prep_group, parent_prep_file_type, signal_nodes, signal_cache,
                          params, failures):
    expression_accs = []
    for node in signal_nodes:
        file_type = node['tag']
        url = node['value']
        nfa = node['nfa'] if 'nfa' in node else None
        data_class = node['dc'] if 'dc' in node else None
        separator = node['ms'] if 'ms' in node else None
        metadata = node.get('metadata', None)
        if file_type in SIGNAL_TAGS:
            signal_group = signal_cache.get(url, None)
            if signal_group is None:
                signal_group = add_signals(file_signal=url,
                                           file_metadata=metadata,
                                           file_type=file_type,
                                           number_of_feature_attributes=nfa,
                                           data_class=data_class,
                                           measurement_separator=separator,
                                           params=params)
                signal_cache[url] = signal_group
            if file_type == 'expression':
                expression_accs.append(signal_group)
            what = '{}_to_{}'.format(file_type, parent_prep_file_type)
            try:
                link_by_parent(what, parent_prep_group, signal_group, params)
            except GroupLinkingError as ex:
                if not params.IGNORE_LINKING_ERRORS:
                    sys.exit(1)
                failures.append(ex)
                continue
        elif file_type == 'mapping-file':
            # mapping file is always the last node, if present
            if is_acc(url):
                map_f_acc = check_mapping_file(url, params)
            else:
                map_f_acc = add_mappings(url, params, metadata)
            for expr_acc in expression_accs:
                link_mappings(map_f_acc, expr_acc, params)


def add_lib_prep(sample_group, nodes, link_cache, params, study, failures):
    for node in nodes:
        lib_prep_file_type = node['tag']
        url = node['value']
        lib_prep_group = link_cache.get(url, None)
        if lib_prep_group is None:
            lib_prep_group = add_and_link_libs_preps(
                sample_group, url, lib_prep_file_type, params, study, failures
            )
            link_cache[url] = lib_prep_group
        children = node.get('children', [])
        add_signals_to_parent(
            lib_prep_group, lib_prep_file_type, children,
            link_cache, params, failures
        )


def existing_lib_prop(node, signal_cache, params, failures):
    lib_prep_file_type = node['tag']
    lib_prep_group = node['value']
    children = node.get('children', [])
    add_signals_to_parent(
        lib_prep_group, lib_prep_file_type, children, signal_cache, params, failures
    )


def handle_lib_prep_case(parser_state, params, study, failures):
    link_cache = {}
    for sample_node in parser_state.sample_node_list:
        value = sample_node['value']
        libraries_and_preparations = sample_node.get('children', [])
        if value == 'implicit':
            assert len(libraries_and_preparations) == 1, \
                'Internal error: expect only one subnode for implicit sample node'
            sub_node = libraries_and_preparations[0]
            existing_lib_prop(sub_node, link_cache, params, failures)
            return

        if is_acc(value):
            # existing samples group, expect some libraries and preparations
            if len(libraries_and_preparations) == 0:
                print('samples argument {} is ignored'.format(value))
                return
            sample_group = value
        else:
            try:
                sample_group = add_and_link_samples(params, study, sample_link=value)
            except GroupLinkingError as ex:
                if not params.IGNORE_LINKING_ERRORS:
                    sys.exit(1)
                failures.append(ex)
                continue

        add_lib_prep(
            sample_group, libraries_and_preparations,
            link_cache, params, study, failures
        )


def handle_samples_signals_case(parser_state, params, study, failures):
    signal_cache = {}
    for sample_node in parser_state.sample_node_list:
        value = sample_node['value']
        signals = sample_node.get('children', [])

        if is_acc(value):
            # existing samples group, expect some signals
            if not signals:
                _err('No signals provided to link, ignoring `--samples {}`'.format(value))
                return
            sample_group = value
        else:
            try:
                sample_group = add_and_link_samples(params, study, sample_link=value)
            except GroupLinkingError as ex:
                if not params.IGNORE_LINKING_ERRORS:
                    sys.exit(1)
                failures.append(ex)
                continue

        add_signals_to_parent(
            sample_group, 'sample', signals, signal_cache, params, failures
        )


def check_for_repeated_links(link, nodes, links_cache):
    links_cache[link] = 'study'
    for node in nodes:
        value = node['value']
        tag = node['tag']
        value = None if value == 'implicit' else value
        if value is not None:
            if value in links_cache:
                _err('Duplicate link or accession: --{} {}'.format(tag, value), in_red=True)
                sys.exit(1)
            links_cache[value] = tag
        meta = node.get('metainfo', None)
        if meta is not None:
            if meta in links_cache:
                _err('Duplicate link or accession: --{} {}'.format(tag, meta), in_red=True)
                sys.exit(1)
            else:
                links_cache[meta] = tag
        children = node.get('children', [])
        check_for_repeated_links(link, children, links_cache)


def collect_all_nodes_by_tag(nodes, filter_tag_fun):
    result = []
    for node in nodes:
        if filter_tag_fun(node['tag']):
            result.append(node)
        children = node.get('children', [])
        result = result + collect_all_nodes_by_tag(children, filter_tag_fun)
    return result


def collect_all_mapping_file_nodes(nodes):
    return collect_all_nodes_by_tag(
        nodes,
        lambda x: x.startswith('mapping-file')
    )


def add_all_signal_args_to_all_lib_preps(sample_nodes):
    all_libs_preps = collect_all_nodes_by_tag(
        sample_nodes,
        lambda x: x in LIB_PREP_TAGS
    )
    all_signal_nodes = collect_all_nodes_by_tag(
        sample_nodes,
        lambda x: x in SIGNAL_TAGS
    )
    mapping_file_nodes = collect_all_mapping_file_nodes(sample_nodes)
    for libs_preps_node in all_libs_preps:
        libs_preps_node['children'] = copy.deepcopy(all_signal_nodes)
    for sample_node in sample_nodes:
        sample_node['children'] = copy.deepcopy(all_libs_preps)
    if mapping_file_nodes:
        sample_node = sample_nodes[0]
        libs_preps_node = sample_node['children'][0]
        libs_preps_node['children'] = all_signal_nodes + mapping_file_nodes


def add_all_signal_args_to_all_samples(sample_nodes):
    all_signal_nodes = collect_all_nodes_by_tag(
        sample_nodes,
        lambda x: x in SIGNAL_TAGS
    )
    for sample_node in sample_nodes:
        children = sample_node.get('children', [])
        non_signal_children = [node for node in children if node['tag'] not in SIGNAL_TAGS]
        sample_node['children'] = all_signal_nodes + non_signal_children


def check_signal_versions(study_acc, parent_accession, parent_tag, signals):
    """ Perform basic validation of signal versioning arguments
    Updating version is allowed when:
    - study already exists
    - samples already exist
    - designated previous version of signal object appears only once in a
      script call
    """
    prev_versions = set()
    for signal_node in signals:
        value = signal_node['value']
        _, prev_version = parse_file_signal(value)
        if prev_version is None:
            continue
        if study_acc is None:
            _err("Previous versions cannot be specified when creating new studies, "
                 "only for signal group objects of existing studies",
                 in_red=True)
            sys.exit(1)
        if parent_accession is None:
            _err("Previous versions cannot be specified when creating new {0}, "
                 "only for signal group objects of existing {0}".format(parent_tag),
                 in_red=True)
            sys.exit(1)
        if prev_version in prev_versions:
            _err("It is not possible to import a new version of an existing "
                 "signal group object version ('{}') more than once "
                 "in the same script call".format(prev_version),
                 in_red=True)
            sys.exit(1)
        prev_versions.add(prev_version)


def check_signal_versions_libs_preps(sample_nodes, study_acc):
    for sample_node in sample_nodes:
        libs_preps = sample_node.get('children', [])
        for libs_preps_node in libs_preps:
            value = libs_preps_node['value']
            parent_accession = value if is_acc(value) else None
            parent_tag = libs_preps_node['tag']
            signals = [node for node in libs_preps_node.get('children', [])
                       if node['tag'] in SIGNAL_TAGS]
            check_signal_versions(study_acc, parent_accession, parent_tag, signals)


def check_signal_versions_samples(sample_nodes, study_acc):
    for sample_node in sample_nodes:
        value = sample_node['value']
        parent_accession = value if is_acc(value) else None
        parent_tag = sample_node['tag']
        signals = [node for node in sample_node.get('children', []) if node['tag'] in SIGNAL_TAGS]
        check_signal_versions(study_acc, parent_accession, parent_tag, signals)


def do_import(import_params):
    parser_args_state = import_params.parser_args_state
    if not (import_params.SERVER.startswith("https://") or import_params.SERVER.startswith("http://")):
        _err("The server url should start with 'http' or 'https'")
        sys.exit(1)
    if not (import_params.study_link or import_params.study_accession):
        _err("No study link or accession was provided", in_red=True)
        sys.exit(1)
    if import_params.study_link and import_params.study_accession:
        _err("You need to provide either study link or accession", in_red=True)
        sys.exit(1)

    check_for_repeated_links(
        import_params.study_link,
        parser_args_state.sample_node_list,
        links_cache={}
    )

    # If the option LINK_SIGNALS_TO_ALL_SAMPLES is set, we try to modify
    # the arguments in such a way as to conform to the behaviour, which is described in
    # ODM-7826.
    # The subsequent code actually works with repeated links for signal data

    if import_params.LINK_SIGNALS_TO_ALL_SAMPLES:
        if parser_args_state.has_libraries_or_preparations():
            add_all_signal_args_to_all_lib_preps(parser_args_state.sample_node_list)
        else:
            add_all_signal_args_to_all_samples(parser_args_state.sample_node_list)

    check_and_merge_mapping_file_in_sample_nodes(parser_args_state.sample_node_list)

    mapping_files_number = len(collect_all_mapping_file_nodes(parser_args_state.sample_node_list))
    if import_params.LINK_SIGNALS_TO_ALL_SAMPLES and mapping_files_number > 1:
        _err('Only one mapping file is expected with link-all-to-all option', in_red=True)
        sys.exit(1)

    if (parser_args_state.has_libraries_or_preparations()
            and parser_args_state.has_non_expression_signals()):
        _err('the linkage between libraries/preparations and variants/flow-cytometry '
             'is not supported', in_red=True)
        sys.exit(1)

    if not parser_args_state.has_consistent_data_model():
        _err('inconsistent data model', in_red=True)
        sys.exit(1)

    if parser_args_state.has_libraries_or_preparations():
        check_signal_versions_libs_preps(
            parser_args_state.sample_node_list,
            import_params.study_accession
        )
    else:
        check_signal_versions_samples(
            parser_args_state.sample_node_list,
            import_params.study_accession
        )

    if import_params.dump_args_as_json:
        print(json.dumps(parser_args_state.sample_node_list, indent=2))
        sys.exit(0)

    study = add_study(params=import_params)
    failures = []
    if parser_args_state.has_libraries_or_preparations():
        handle_lib_prep_case(parser_args_state, import_params, study, failures)
    else:
        handle_samples_signals_case(parser_args_state, import_params, study, failures)

    print(green_text(u"Execution is finished!"))
    if failures:
        sys.exit(1)
    return study


class ImportParams:
    # The class constructor is a subject to change, its parameters might be changed in the future.
    def __init__(
            self,
            server,
            headers=None,
            token=None,
            study_link=None,
            study_accession=None,
            app_version="default-released",
            etl_source=None,
            template_accession_supplier=lambda: None,
            allow_duplicates=False,
            number_of_feature_attributes=None,
            data_class=None,
            measurement_separator=None,
            fail_if_file_exists=False,
            link_signals_to_all_samples=False,
            ignore_linking_errors=False,
            job_timeout=ETL_WAITING_TIMEOUT,
            debug=False,
            dump_args_as_json=False,
            # provide dedicated links to entities
            # or parser_args_state, containing any of them
            # (used for command line call, wider functionality is supported)
            samples_link=None,
            libraries_link=None,
            preparations_link=None,
            expression_link=None,
            expression_metadata_link=None,
            variant_link=None,
            variant_metadata_link=None,
            flow_cytometry_link=None,
            flow_cytometry_metadata_link=None,
            parser_args_state=None
    ):
        self.SERVER = server.rstrip('/')
        self.study_link = study_link
        self.study_accession = study_accession
        self.APP_VERSION = app_version
        self.ETL_SOURCE = etl_source
        self.TEMPLATE_ACCESSION_SUPPLIER = template_accession_supplier
        self.ALLOW_DUPLICATES = allow_duplicates
        self.NUMBER_OF_FEATURE_ATTRIBUTES = number_of_feature_attributes
        self.DATA_CLASS = data_class
        self.MEASUREMENT_SEPARATOR = measurement_separator
        self.FAIL_IF_FILE_EXISTS = fail_if_file_exists
        self.LINK_SIGNALS_TO_ALL_SAMPLES = link_signals_to_all_samples
        self.IGNORE_LINKING_ERRORS = ignore_linking_errors
        self.JOB_TIMEOUT = job_timeout
        self.debug = debug
        self.dump_args_as_json = dump_args_as_json
        if headers is not None:
            self.headers = headers
        elif token is not None:
            self.headers = {"Genestack-API-Token": token,
                            "Accept": "application/json",
                            "Content-Type": "application/json"}
        else:
            _err("You should either provide headers or genestack api token")
            sys.exit(1)
        if not parser_args_state:
            parser_args_state = ImportParams._fill_parser_from_args(
                samples_link, libraries_link, preparations_link,
                expression_link, expression_metadata_link,
                variant_link, variant_metadata_link,
                flow_cytometry_link, flow_cytometry_metadata_link
            )
        self.parser_args_state = parser_args_state

    @staticmethod
    def _fill_parser_from_args(
            samples_link,
            libraries_link,
            preparations_link,
            expression_link,
            expression_metadata_link,
            variant_link,
            variant_metadata_link,
            flow_cytometry_link,
            flow_cytometry_metadata_link
    ):
        parser_args_state = ParserAstState()
        if samples_link:
            make_samples_action(parser_args_state)(
                dest="data",
                option_strings=[]
            ).handle_action(
                tag="samples",
                value=samples_link,
                option_string=None
            )
        if libraries_link:
            make_libraries_and_preparations_action(parser_args_state)(
                dest="data",
                option_strings=[]
            ).handle_action(
                tag="libraries",
                value=libraries_link,
                option_string=None
            )
        if preparations_link:
            make_libraries_and_preparations_action(parser_args_state)(
                dest="data",
                option_strings=[]
            ).handle_action(
                tag="preparations",
                value=preparations_link,
                option_string=None
            )
        if expression_link:
            make_signal_action(parser_args_state)(
                dest="data",
                option_strings=[]
            ).handle_action(
                tag="expression",
                value=expression_link,
                option_string=None
            )
        if expression_metadata_link:
            make_signal_action(parser_args_state)(
                dest="data",
                option_strings=[]
            ).handle_action(
                tag="expression-metadata",
                value=expression_metadata_link,
                option_string=None
            )
        if variant_link:
            make_signal_action(parser_args_state)(
                dest="data",
                option_strings=[]
            ).handle_action(
                tag="variant",
                value=variant_link,
                option_string=None
            )
        if variant_metadata_link:
            make_signal_action(parser_args_state)(
                dest="data",
                option_strings=[]
            ).handle_action(
                tag="variant-metadata",
                value=variant_metadata_link,
                option_string=None
            )
        if flow_cytometry_link:
            make_signal_action(parser_args_state)(
                dest="data",
                option_strings=[]
            ).handle_action(
                tag="flow-cytometry",
                value=flow_cytometry_link,
                option_string=None
            )
        if flow_cytometry_metadata_link:
            make_signal_action(parser_args_state)(
                dest="data",
                option_strings=[]
            ).handle_action(
                tag="flow-cytometry-metadata",
                value=flow_cytometry_metadata_link,
                option_string=None
            )
        return parser_args_state

    @classmethod
    def from_parsed_params(cls, args, parser_args_state):
        return cls(
            server=args.SERVER,
            headers=Headers(args),
            study_link=args.study_link,
            study_accession=args.study_accession,
            app_version=args.APP_VERSION,
            etl_source=args.ETL_SOURCE,
            template_accession_supplier=TemplateAccessionSupplier(args),
            allow_duplicates=args.ALLOW_DUPLICATES,
            number_of_feature_attributes=args.NUMBER_OF_FEATURE_ATTRIBUTES,
            data_class=args.DATA_CLASS,
            measurement_separator=args.MEASUREMENT_SEPARATOR,
            fail_if_file_exists=args.FAIL_IF_FILE_EXISTS,
            link_signals_to_all_samples=args.LINK_SIGNALS_TO_ALL_SAMPLES,
            ignore_linking_errors=args.IGNORE_LINKING_ERRORS,
            job_timeout=args.JOB_TIMEOUT,
            debug=args.debug,
            dump_args_as_json=args.dump_args_as_json,
            parser_args_state=parser_args_state
        )


def main():
    parser_args_state = ParserAstState()
    parser = SaneArgumentParser(description=__doc__,
                                epilog="Old style options using underscores "
                                       "(e.g., '--link_all_to_all') are still supported, "
                                       "but considered obsolete",
                                formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument("-t", "--token",
                        action=prevent_redundant_parameters(("API_TOKEN", "ACCESS_TOKEN"),
                                                        "only one token can be specified. "
                                                        "Please choose the authentication method: "
                                                        "through the Access Token or "
                                                        "through the Genestack-API-token"),
                        dest="API_TOKEN",
                        nargs="?",
                        help="API_TOKEN")
    parser.add_argument("-at", "--access-token",
                        action=prevent_redundant_parameters(("API_TOKEN", "ACCESS_TOKEN"),
                                                        "only one token can be specified. "
                                                        "Please choose the authentication method: "
                                                        "through the Access Token or "
                                                        "through the Genestack-API-token"),
                        dest="ACCESS_TOKEN",
                        nargs="?",
                        help="ACCESS_TOKEN")
    parser.add_argument("-H", "--host", "-srv", "--server",
                        action=_StoreServerName,
                        const="https://odm-demos.genestack.com/",
                        default="https://odm-demos.genestack.com/",
                        nargs="?",
                        dest="SERVER",
                        help="URL of the instance data is being loaded to")
    parser.add_argument("-tmpl", "--template",
                        action="store",
                        const=None,
                        default=None,
                        nargs="?",
                        dest="TEMPLATE_ACCESSION",
                        help="Accession of the template, by default will use "
                             "template marked as default.")
    parser.add_argument("--debug",
                        action="store",
                        const=True,
                        default=False,
                        nargs="?",
                        dest="debug",
                        help="Enable debug mode.")
    parser.add_argument("-s", "--study",
                        action=prevent_redundant_parameters(('study_link', 'study_accession'),
                            "provide exactly one study via a link or an existing study accession"),
                        dest="study_link",
                        help="link to study file")
    parser.add_argument("-sa", "--study-accession",
                        action=prevent_redundant_parameters(('study_link', 'study_accession'),
                            "provide exactly one study via a link or an existing study accession"),
                        dest="study_accession",
                        help="accession of existing study")
    parser.add_argument("-sm", "--samples",
                        action=make_samples_action(parser_args_state),
                        dest="data",
                        metavar="SAMPLES_LINK_OR_ACCESSION",
                        nargs="?",
                        help="existing sample group accession "
                             "or link to file with sample data")
    parser.add_argument("-la", "--link-attribute",
                        # "Attribute to link data with sample source"
                        # linking by group does not support custom linking attribute
                        # it's left for compatibility reasons
                        action=DeprecatedAction,
                        help=argparse.SUPPRESS)
    parser.add_argument("-lb", "--libraries",
                        action=make_libraries_and_preparations_action(parser_args_state),
                        dest="data",
                        metavar="LIBRARIES_LINK_OR_ACCESSION",
                        help="link to libraries data file or accession of existing library group",
                        nargs="?")
    parser.add_argument("-pr", "--preparations",
                        action=make_libraries_and_preparations_action(parser_args_state),
                        dest="data",
                        metavar="PREPARATIONS_LINK_OR_ACCESSION",
                        help="link to preparations data file or accession of "
                             "existing preparation group",
                        nargs="?")
    parser.add_argument("-e", "--expression",
                        action=make_signal_action(parser_args_state),
                        dest="data",
                        metavar="SIGNAL_LINK",
                        help="link to tabular data file (not only expression data) except Gene "
                             "Variant or Flow Cytometry",
                        nargs="?")
    parser.add_argument("-em", "--expression-metadata",
                        action=make_signal_action(parser_args_state),
                        dest="data",
                        metavar="SIGNAL_METADATA_LINK",
                        help="link to tabular metadata file (not only expression data) except Gene "
                             "Variant or Flow Cytometry",
                        nargs="?")
    parser.add_argument("-v", "--variant",
                        action=make_signal_action(parser_args_state),
                        dest="data",
                        metavar="SIGNAL_LINK",
                        help="link to variants data file",
                        nargs="?")
    parser.add_argument("-vm", "--variant-metadata",
                        action=make_signal_action(parser_args_state),
                        dest="data",
                        metavar="SIGNAL_METADATA_LINK",
                        help="link to variant metadata file",
                        nargs="?")
    parser.add_argument("-f", "--flow-cytometry",
                        action=make_signal_action(parser_args_state),
                        dest="data",
                        metavar="SIGNAL_LINK",
                        help="link to flow cytometry data file",
                        nargs="?")
    parser.add_argument("-fm", "--flow-cytometry-metadata",
                        action=make_signal_action(parser_args_state),
                        dest="data",
                        metavar="SIGNAL_METADATA_LINK",
                        help="link to flow cytometry metadata file",
                        nargs="?")
    parser.add_argument("-nfa", "--number-of-feature-attributes",
                        action=make_signal_action(parser_args_state),
                        dest="NUMBER_OF_FEATURE_ATTRIBUTES",
                        metavar="NUMBER_OF_FEATURE_ATTRIBUTES",
                        help="number of feature attribute columns in a tsv file",
                        nargs="?")
    parser.add_argument("-dc", "--data-class",
                        action=make_signal_action(parser_args_state),
                        dest="DATA_CLASS",
                        metavar="DATA_CLASS",
                        help="describes the data type of the uploaded file",
                        nargs="?")
    parser.add_argument("-ms", "--measurement-separator",
                        action=make_signal_action(parser_args_state),
                        dest="MEASUREMENT_SEPARATOR",
                        metavar="MEASUREMENT_SEPARATOR",
                        help="separator for sample/library/preparation name from the measurement name in column names in a tsv file",
                        nargs="?")
    parser.add_argument("-mpf", "--mapping-file",
                        action=make_mapping_file_action(parser_args_state),
                        dest="data",
                        metavar="MAPPING_FILE_LINK",
                        help="link to mapping file",
                        nargs="?")
    parser.add_argument("-mpfa", "--mapping-file-accession",
                        action=make_mapping_file_action(parser_args_state),
                        dest="data",
                        metavar="MAPPING_FILE_ACCESSION",
                        help="accession of the existing mapping file",
                        nargs="?")
    parser.add_argument("-mpfm", "--mapping-file-metadata",
                        action=make_mapping_file_action(parser_args_state),
                        dest="data",
                        metavar="MAPPING_FILE_METADATA_LINK",
                        help="link to metadata file for this mapping file",
                        nargs="?")
    parser.add_argument("-av", "--app-version",
                        action="store",
                        default="default-released",
                        dest="APP_VERSION",
                        help="application version, appended to server URL to select "
                             "a particular endpoint version; default is '%(default)s' "
                             "(it is recommended not to change this parameter)",
                        nargs="?")
    parser.add_argument("-T", "--job-timeout", dest="JOB_TIMEOUT",
                        metavar="TIMEOUT-SECONDS", type=int, default=ETL_WAITING_TIMEOUT,
                        help="time to wait for import jobs to finish (in seconds), "
                             "defaults to %(default)s")
    parser.add_argument("--import-source", dest="ETL_SOURCE", metavar="PROTOCOL",
                        choices=ETL_SOURCES,
                        help="protocol (source) for server to fetch data; valid choices are: {}\n"
                             "Usually not required, but you might need to override "
                             "heuristics based on URL itself and server configuration"
                             "".format(ETL_SOURCES))
    parser.add_argument("-lata", "--link-all-to-all",
                        dest="LINK_SIGNALS_TO_ALL_SAMPLES",
                        action="store_true",
                        default=False,
                        help="Link all signals to all samples, instead of "
                             "preceding 'samples' file only")
    parser.add_argument("-ile", "--ignore-linking-errors",
                        dest="IGNORE_LINKING_ERRORS",
                        action="store_true",
                        default=False,
                        help="To allow the script to continue even if errors "
                             "linking study, samples and signal files occur")
    parser.add_argument('-J', '--dump-args-as-json',
                        action="store_true",
                        default=False,
                        help=argparse.SUPPRESS)
    behaviour_args = parser.add_mutually_exclusive_group()
    behaviour_args.add_argument("--allow-duplicates", dest="ALLOW_DUPLICATES",
                                action='store_true',
                                help="allow uploading of duplicate data (warning: "
                                     "this is *not* recommended outside of training use or tests)")
    behaviour_args.add_argument("--fail-if-exists", dest="FAIL_IF_FILE_EXISTS",
                                action='store_true',
                                help=("exit if the file being imported already exists on the "
                                      "server (default is to re-use existing file)"))
    args = parser.parse_args()

    do_import(ImportParams.from_parsed_params(args, parser_args_state))


if __name__ == "__main__":
    main()



from copy import deepcopy

from odm_sdk import FileTypes, GenestackException, Permissions, validate_constant


class FileFilter(object):
    """
    Base file filter class.
    """
    def __init__(self):
        self._dict = {}

    def get_dict(self):
        return deepcopy(self._dict)

    def AND(self, other):
        """
        Return a new filter combining this one with another one in an AND clause.

        :param other: other filter
        :type other: FileFilter
        :rtype: FileFilter
        """
        return AndFileFilter(self, other)

    def OR(self, other):
        """
        Return a new filter combining this one with another one in an OR clause.

        :param other: other filter
        :type other: FileFilter
        :rtype: FileFilter
        """
        return OrFileFilter(self, other)

    def __and__(self, other):
        return self.AND(other)

    def __or__(self, other):
        return self.OR(other)


class TypeFileFilter(FileFilter):
    """
    Filter to select files with a given file type.
    See :ref:`fileTypes` for a list of possible file types.
    """
    def __init__(self, file_type):
        super(TypeFileFilter, self).__init__()
        if not validate_constant(FileTypes, file_type):
            raise GenestackException("Invalid file type")
        self._dict.update({'type': file_type})


class KeyValueFileFilter(FileFilter):
    """
    Filter to select files with a given metainfo key-value pair.
    """
    def __init__(self, key, value):
        super(KeyValueFileFilter, self).__init__()
        self._dict.update({'keyValue': {'key': key, 'value': value}})


class OwnerFileFilter(FileFilter):
    """
    Filter to select files owned by a specific user.
    """
    def __init__(self, email):
        super(OwnerFileFilter, self).__init__()
        self._dict.update({'owner': email})


class MetainfoValuePatternFileFilter(FileFilter):
    """
    Filter to select files matching a specific substring value for a metainfo key.
    """
    def __init__(self, key, value):
        super(MetainfoValuePatternFileFilter, self).__init__()
        self._dict.update({'pattern': {'key': key, 'value': value}})


class ChildrenFileFilter(FileFilter):
    """
    Filter to select files that are the children or descendants of a given container.
    """
    def __init__(self, container, recursive=False):
        super(ChildrenFileFilter, self).__init__()
        self._dict.update({'children': {'file': container, 'recursive': recursive}})


class ContainsFileFilter(FileFilter):
    """
    Filter to select containers that contain a given file.
    """
    def __init__(self, file_accession):
        super(ContainsFileFilter, self).__init__()
        self._dict.update({'contains': file_accession})


class ActualOwnerFileFilter(FileFilter):
    """
    Filter to select files that are owned by the current user.
    """
    def __init__(self):
        super(ActualOwnerFileFilter, self).__init__()
        self._dict.update({'owned': None})


class BelongsToDatasetFileFilter(FileFilter):
    """
    Same as :py:class:`~odm_sdk.file_filters.ChildrenFileFilter`
    but searches for files that belong to the specified dataset.
    """
    def __init__(self, file_accession):
        super(BelongsToDatasetFileFilter, self).__init__()
        self._dict.update({'datasetAccession': file_accession})


class ActualPermissionFileFilter(FileFilter):
    """
    Filter to select files for which the current user has a specific permission.
    See :ref:`permissions`.
    """
    def __init__(self, permission):
        super(ActualPermissionFileFilter, self).__init__()
        if not validate_constant(Permissions, permission):
            raise GenestackException("Invalid permission")
        self._dict.update({'access': permission})


class FixedValueFileFilter(FileFilter):
    """
    Fixed value filter (either ``True`` or ``False``).
    """
    def __init__(self, value):
        super(FixedValueFileFilter, self).__init__()
        self._dict.update({'fixed': value})


class HasInProvenanceFileFilter(FileFilter):
    """
    Filter to select files that have a given file in their provenance graph.
    """
    def __init__(self, file_accession):
        super(HasInProvenanceFileFilter, self).__init__()
        self._dict.update({'hasInProvenance': file_accession})


class PermissionFileFilter(FileFilter):
    """
    Filter to select files for which a specific group has a specific permission.
    See :ref:`permissions`.
    """
    def __init__(self, group, permission):
        super(PermissionFileFilter, self).__init__()
        if not validate_constant(Permissions, permission):
            raise GenestackException("Invalid permission")
        self._dict.update({'permission': {'group': group, 'value': permission}})


class NotFileFilter(FileFilter):
    """
    Negation of another :py:class:`~odm_sdk.file_filters.FileFilter`
    """
    def __init__(self, other_filter):
        super(NotFileFilter, self).__init__()
        self._dict.update({'not': other_filter.get_dict()})


class AndFileFilter(FileFilter):
    """
    "AND" combination of two file filters.
    """
    def __init__(self, first, second):
        super(AndFileFilter, self).__init__()
        self._dict.update({'and': [first.get_dict(), second.get_dict()]})


class OrFileFilter(FileFilter):
    """
    "OR" combination of two file filters.
    """
    def __init__(self, first, second):
        super(OrFileFilter, self).__init__()
        self._dict.update({'or': [first.get_dict(), second.get_dict()]})



import os
import re
import sys
from argparse import ArgumentParser
from getpass import getpass
from operator import attrgetter

from odm_sdk import GenestackAuthenticationException
from odm_sdk.shell import Command, GenestackShell
from odm_sdk.settings import User, config
from odm_sdk.utils import interactive_select


def input_host():
    print('Please input Genestack host address')
    while True:
        host = input('host: ').strip()
        if not host:
            print('Host can not be empty')
            continue
        return host


def validate_alias(alias):
    expression = re.compile('[a-zA-Z0-9_@\-]+$')
    return bool(alias and expression.match(alias))


def input_alias(existing):
    print('Please input alias. (Alias can contain: letters (a-z, A-Z), '
          'digits (0-9), at-sign (@), underscore (_), hyphen (-))')
    while True:
        alias = input('alias: ').strip()
        if not alias:
            print('Alias cannot be empty')
            continue
        if not validate_alias(alias):
            print('Restricted symbols message')
            continue
        if alias in existing:
            print('Alias must be unique')
            continue
        return alias


def create_user_from_input(host, alias):
    """
    Ask credentials interactively and return user that can login to platform.

    :param host:  server host
    :type host:  basestring
    :param alias: user alias
    :type alias: basestring
    :return: user
    :rtype: User
    """
    by_token = 'by token'
    by_access_token = 'by access token'
    items = [by_token, by_access_token, 'by email and password']
    choice = interactive_select(items, 'Select authentication')

    if choice == by_token:
        return create_user_from_token(host, alias=alias)
    elif choice == by_access_token:
        return create_user_from_access_token(host, alias=alias)
    else:
        return create_user_from_input_email_and_password(host, alias=alias)


def create_user_from_input_email_and_password(host, alias=None):
    """
    Ask email and password, check that it is possible to login with this credentials
    and return user.

    :param host:  server host
    :type host:  basestring
    :param alias: user alias
    :type alias: basestring
    :return: user
    :rtype: User
    """
    print('Specify email and password for host: "%s"' % host, end=' ')
    if alias:
        print('and alias: "%s"' % alias)
    else:
        print()
    user_login = None
    while True:
        if user_login:
            res = input('Please specify your user login (email) [%s]: ' % user_login).strip()
            if res:
                user_login = res
        else:
            user_login = input('Please specify your user login (email): ').strip()
            if not user_login:
                print('Login cannot be empty')
                continue
        user_password = getpass('Please specify your password for %s: ' % user_login)
        if not user_password:
            print('Password cannot be empty')
            continue

        if not user_login or not user_password:
            print()
            continue
        user = User(user_login, host=host, password=user_password, alias=alias)
        try:
            user.get_connection()
            break
        except GenestackAuthenticationException:
            print('Your username or password was incorrect, please try again')
    return user


def create_user_from_token(host, alias=None):
    print('Host: %s' % host)
    msg = 'Please specify Genestack API token%s: '
    with_alias = '' if not alias else ' for "%s"' % alias
    msg = msg % with_alias
    while True:
        token = getpass(msg)
        if not token:
            print('Token cannot be empty')
            continue
        user = User(host=host, alias=alias, token=token)
        try:
            user.get_connection()
            break
        except GenestackAuthenticationException:
            print('Could not login with given token, please try again')
    return user


def create_user_from_access_token(host, alias=None):
    print(f'Host: {host}')
    with_alias = f' for "{alias}"' if alias else ''
    msg = f'Please specify access token or environment variable with its value{with_alias}: '
    while True:
        access_token = input(msg)
        if not access_token:
            print('Token cannot be empty')
            continue
        user = User(host=host, alias=alias, access_token=access_token)
        try:
            user.get_connection()
            break
        except GenestackAuthenticationException:
            print('Could not login with given token, please try again')
    return user


def check_config():
    config_path = config.get_settings_file()
    if not os.path.exists(config_path):
        print('You do not seem to have a config file yet. '
              'Please run `odm-user-setup init`. Exiting')
        exit(1)


class AddUser(Command):
    COMMAND = 'add'
    DESCRIPTION = 'Add new user.'
    OFFLINE = True

    def run(self):
        alias = input_alias(config.users.keys())
        host = input_host()
        user = create_user_from_input(host, alias)
        config.add_user(user)
        print('User "%s" has been created' % user.alias)


def select_user(users, selected=None):
    """
    Choose user from users stored in config.

    :param users:
    :param selected:
    :return:
    :rtype: User
    """
    user_list = sorted(users.values(), key=lambda x: x.alias)
    return interactive_select(user_list, 'Select user', to_string=attrgetter('alias'), selected=selected)


class ChangePassword(Command):
    COMMAND = 'change-password'
    DESCRIPTION = 'Change password for user.'
    OFFLINE = True

    def update_parser(self, parent):
        parent.add_argument('alias', metavar='<alias>', help='Alias for user to change password', nargs='?')

    def run(self):
        check_config()
        users = config.users
        user = users.get(self.args.alias)
        if not user:
            user = select_user(users)

        if not user.email:
            print('User without email could be authorized only by token')
            return

        while True:
            user.password = getpass('Input password for %s: ' % user.alias.encode('utf-8'))
            try:
                user.get_connection()
                break
            except GenestackAuthenticationException:
                continue
        config.change_password(user.alias, user.password)
        print('Password has been changed successfully')


class ChangeToken(Command):
    COMMAND = 'change-token'
    DESCRIPTION = 'Change token for user.'
    OFFLINE = True

    def update_parser(self, parent):
        parent.add_argument('alias', metavar='<alias>',
                            help='Alias for user to change token for', nargs='?')

    def run(self):
        check_config()
        users = config.users
        user = users.get(self.args.alias)
        if not user:
            user = select_user(users)
        new_user = create_user_from_token(user.host, alias=user.alias)
        user.token = new_user.token
        config.change_token(user.alias, user.token)
        print('Token has been changed successfully')


class SetDefault(Command):
    COMMAND = 'default'
    DESCRIPTION = 'Set default user.'
    OFFLINE = True

    def update_parser(self, parent):
        parent.add_argument('alias', metavar='<alias>', help='Alias for user to change password', nargs='?')

    def run(self):
        check_config()
        users = config.users
        user = users.get(self.args.alias)
        if not user:
            user = select_user(users, selected=config.default_user)
        if user.alias != config.default_user.alias:
            config.set_default_user(user)
            print('Default user has been set to "%s"' % user.alias)
        else:
            print('Default user has not been changed')


class Remove(Command):
    COMMAND = 'remove'
    DESCRIPTION = 'Remove user.'
    OFFLINE = True

    def update_parser(self, parent):
        parent.add_argument('alias', metavar='<alias>', help='Alias for user to change password', nargs='?')

    def run(self):
        check_config()
        users = config.users

        user = users.get(self.args.alias)
        if not user:
            user = select_user(users)
        if user.alias == config.default_user.alias:
            print('Cannot delete default user')
            return
        config.remove_user(user)
        print('"%s" has been removed from config' % user.alias)


class RenameUser(Command):
    COMMAND = 'rename'
    DESCRIPTION = 'Rename user.'
    OFFLINE = True

    def update_parser(self, parent):
        parent.add_argument('alias', metavar='<alias>', help='Alias to be renamed', nargs='?')
        parent.add_argument('new_alias', metavar='<new_alias>', help='New alias', nargs='?')

    def run(self):
        check_config()
        users = config.users

        user = users.get(self.args.alias)

        if not user:
            print('Select user to rename')
            user = select_user(users)
        if not self.args.new_alias or not validate_alias(self.args.new_alias):
            print('Enter new alias')
            new_alias = input_alias(users.keys())
        else:
            new_alias = self.args.new_alias

        new_user = User(email=user.email, alias=new_alias, host=user.host, password=user.password,
                        token=user.token)

        config.add_user(new_user, save=False)
        if user.alias == config.default_user.alias:
            config.set_default_user(new_user, save=False)

        config.remove_user(user)
        print('"%s" alias changed to "%s"' % (user.alias, new_user.alias))


class List(Command):
    COMMAND = 'list'
    DESCRIPTION = 'List all users.'
    OFFLINE = True

    def run(self):
        check_config()
        users = sorted(config.users.items())

        default_user_alias = config.default_user and config.default_user.alias

        for key, user in users:
            print()
            print('%s%s:' % (key, ' (default)' if default_user_alias == key else ''))
            print('  %-10s%s' % ('email', user.email))
            print('  %-10s%s' % ('host', user.host))


class Path(Command):
    COMMAND = 'path'
    DESCRIPTION = 'Show path to configuration file.'
    OFFLINE = True

    def run(self):
        print(config.get_settings_file())


class Init(Command):
    COMMAND = 'init'
    DESCRIPTION = 'Create default settings.'
    OFFLINE = True

    def get_command_parser(self, parser=None):
        parser = parser or ArgumentParser(description=self.DESCRIPTION)
        parser.description = self.DESCRIPTION
        group = parser.add_argument_group('command arguments')
        self.update_parser(group)
        group.add_argument('-H', '--host',
                           help='ODM host address',
                           metavar='<host>')
        return parser

    def run(self):
        """
        Create config file if it is not present.

        Catch ``KeyboardInterrupt`` and ``EOFError`` is required here for case
        when this command is run for first time and in shell mode.
        If we don't quit here, shell will continue execution and ask credentials once more.
        """

        # Hardcoded alias that created for the first user only.
        # Normal usecase is when user have single account and don't care about alias name.
        # Advanced users can rename alias.
        default_alias = 'Default'

        try:
            config_path = config.get_settings_file()
            if os.path.exists(config_path):
                print('A config file already exists at %s' % config_path)
                return
            print('If you do not have a ODM account, you need to create one first')

            user = create_user_from_input(self.args.host, default_alias)
            config.add_user(user)  # adding first user make him default.
            print('Config file at "%s" has been created successfully' % config_path)
        except (KeyboardInterrupt, EOFError):
            sys.stdout.flush()
            sys.stderr.write('\nError: Init is not finished\n')
            exit(1)


class UserManagement(GenestackShell):
    DESCRIPTION = 'Genestack user management application.'
    COMMAND_LIST = [
        Init,
        List,
        AddUser,
        SetDefault,
        ChangePassword,
        ChangeToken,
        Path,
        Remove,
        RenameUser
    ]
    intro = "User setup shell.\nType 'help' for list of available commands.\n\n"
    prompt = 'user_setup> '

    def set_shell_user(self, args):
        if not args.host:
            sys.stderr.write("ODM host is mandatory. Please specify the host using the -H flag\n")
            sys.exit(1)

        config_path = config.get_settings_file()
        if not os.path.exists(config_path):
            print('No config file was found, creating one interactively')
            self.process_command(Init(), ['--host', args.host], False)
            args.host = None  # do not provide host for future use of arguments


def main():
    shell = UserManagement()
    shell.cmdloop()


if __name__ == '__main__':
    main()



import sys
import unittest
from contextlib import contextmanager
from io import StringIO

import requests_mock

from odm_sdk.scripts.import_ODM_data import ImportParams, do_import

JOB_API_PATH = "frontend/rs/genestack/job/default-released"
INTEGRATION_LINK_PATH = "frontend/rs/genestack/integrationCurator/default-released/integration/link"


@contextmanager
def captured_output():
    new_out, new_err = StringIO(), StringIO()
    old_out, old_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = new_out, new_err
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout, sys.stderr = old_out, old_err


class ImportTest(unittest.TestCase):

    def test_exception_on_both_study_identifiers(self):
        with self.assertRaises(SystemExit) as cm, captured_output() as (out, err):
            do_import(ImportParams(
                server="https://dummy",
                headers={"Authenticate": "Token example"},
                study_link="provided link",
                study_accession="provided accession",
            ))
        self.assertEqual(1, cm.exception.code)
        error_text = err.getvalue().strip()
        self.assertEqual('You need to provide either study link or accession', error_text)

    def test_exception_on_providing_no_tokens(self):
        with self.assertRaises(SystemExit) as cm, captured_output() as (out, err):
            args = _MockArgs({
                "SERVER": "https://dummy",
                "study_link": "https://provided.link"
            })
            ImportParams.from_parsed_params(args, None)
        self.assertEqual(1, cm.exception.code)
        error_text = err.getvalue().strip()
        self.assertEqual('Please provide a Genestack-Api-Token or an Access Token', error_text)

    def test_exception_on_providing_two_tokens(self):
        with self.assertRaises(SystemExit) as cm, captured_output() as (out, err):
            args = _MockArgs({
                "ACCESS_TOKEN": "aToken",
                "API_TOKEN": "bToken",
                "SERVER": "https://dummy",
                "study_link": "https://provided.link"
            })
            ImportParams.from_parsed_params(args, None)
        self.assertEqual(1, cm.exception.code)
        error_text = err.getvalue().strip()
        self.assertEqual('Please provide a Genestack-Api-Token or an Access Token but not both',
                         error_text)

    def test_exception_on_providing_api_token(self):
        args = _MockArgs({
            "API_TOKEN": "bToken",
            "TEMPLATE_ACCESSION": "some",
            "SERVER": "https://dummy",
            "study_link": "https://provided.link"
        })
        ImportParams.from_parsed_params(args, None)

    def test_exception_on_providing_access_token(self):
        args = _MockArgs({
            "ACCESS_TOKEN": "aToken",
            "TEMPLATE_ACCESSION": "some",
            "SERVER": "https://dummy",
            "study_link": "https://provided.link"
        })
        ImportParams.from_parsed_params(args, None)

    def test_exception_on_providing_access_token_wo_accession(self):
        with self.assertRaises(SystemExit) as cm, captured_output() as (out, err):
            args = _MockArgs({
                "ACCESS_TOKEN": "aToken",
                "SERVER": "https://dummy",
                "study_link": "https://provided.link"
            })
            ImportParams.from_parsed_params(args, None)
        self.assertEqual(1, cm.exception.code)
        error_text = err.getvalue().strip()
        self.assertEqual('Please provide template accession when using Access Token', error_text)

    def test_ImportParams_constructed_with_token(self):
        ip = ImportParams(server="https://dummy",
                          token="aToken",
                          study_link="provided link",
                          study_accession="provided accession")
        self.assertEqual("aToken", ip.headers["Genestack-API-Token"])

    def test_ImportParams_do_not_constructed_wo_headers_and_token(self):
        with self.assertRaises(SystemExit) as cm, captured_output() as (out, err):
            ImportParams(server="https://wow")
        self.assertEqual(1, cm.exception.code)
        error_text = err.getvalue().strip()
        self.assertEqual('You should either provide headers or genestack api token',
                         error_text)

    @requests_mock.Mocker()
    def test_linking_error_is_visible_to_a_user(self, m):
        srv = "https://dummy.genestack.com"

        # mock study import
        study_job_id = 1
        study_job = {
            "jobExecId": study_job_id,
            "startedBy": "Genestack Superuser",
            "jobName": "IMPORT_STUDY_TSV",
            "status": "STARTING",
            "createTime": "16-11-2022 08:40:14"
        }
        m.post(f"{srv}/{JOB_API_PATH}/import/study/", json=study_job)
        study_job["status"] = "COMPLETED"
        m.get(f"{srv}/{JOB_API_PATH}/{study_job_id}/info", json=study_job)
        study_accession = "GSF020175"
        study_job_result = {
            "result": {
                "accession": study_accession
            }
        }
        m.get(f"{srv}/{JOB_API_PATH}/{study_job_id}/output", json=study_job_result)

        # mock samples import
        samples_job_id = 2
        samples_job = {
            "jobExecId": samples_job_id,
            "startedBy": "Genestack Superuser",
            "jobName": "IMPORT_SAMPLES_TSV",
            "status": "STARTING",
            "createTime": "16-11-2022 08:41:14"
        }
        m.post(f"{srv}/{JOB_API_PATH}/import/samples/", json=samples_job)
        samples_job["status"] = "COMPLETED"
        m.get(f"{srv}/{JOB_API_PATH}/{samples_job_id}/info", json=samples_job)
        samples_group_accession = "GSF020176"
        samples_group_result = {
            "result": {
                "groupAccession": samples_group_accession
            }
        }
        m.get(f"{srv}/{JOB_API_PATH}/{samples_job_id}/output", json=samples_group_result)

        # mock linking - gateway time-out
        m.post(f"{srv}/{INTEGRATION_LINK_PATH}"
               f"/sample/group/{samples_group_accession}/to/study/{study_accession}",
               text="""<html>
                           <head><title>504 Gateway Time-out</title></head>
                           <body bgcolor="white">
                               <center><h1>504 Gateway Time-out</h1></center>
                               <hr><center>nginx</center>
                           </body>
                       </html>""",
               status_code=504)

        with self.assertRaises(SystemExit) as cm, captured_output() as (out, err):
            do_import(ImportParams(
                server=srv,
                headers={"Genestack-API-Token": "aToken"},
                study_link="s3://dummy-bucket/dummy-path/study.tsv",
                samples_link="s3://dummy-bucket/dummy-path/samples.tsv"
            ))
        self.assertEqual(1, cm.exception.code)
        study_import_output, samples_import_output = out.getvalue().splitlines()
        self.assertEqual(f"study {study_accession} was added successfully", study_import_output)
        self.assertEqual(f"samples were added successfully "
                         f"(sample group accession is {samples_group_accession})",
                         samples_import_output)
        linking_error_message, linking_server_response = err.getvalue().split('\n', 1)
        self.assertEqual(f"Linking samples to study "
                         f"({samples_group_accession} to {study_accession}) failed",
                         linking_error_message)
        self.assertIn("504 Gateway Time-out", linking_server_response)


class _MockArgs(dict):
    def __getattr__(self, item):
        try:
            return self[item]
        except KeyError:
            return None


if __name__ == '__main__':
    unittest.main()



from odm_sdk import (Application, FileFilter, GenestackException, Metainfo,
                              validate_constant)

CALCULATE_CHECKSUMS_KEY = 'genestack.checksum:markedForTests'
EXPECTED_CHECKSUM_PREFIX = 'genestack.checksum.expected:'
FILE_BATCH_SIZE = 500


class SpecialFolders(object):
    """
    - ``IMPORTED``: folder with files created by Data Importers
    - ``CREATED``: folder with files created by ``Preprocess`` and ``Analyse`` applications
    - ``TEMPORARY``: folder with temporary files
    - ``UPLOADED``: folder with uploaded raw files
    - ``MY_DATASETS``: folder with created datasets
    """
    IMPORTED = 'imported'
    CREATED = 'created'
    TEMPORARY = 'temporary'
    UPLOADED = 'uploaded'
    MY_DATASETS = 'my datasets'


class SortOrder(object):
    """
    Sort orders for file search queries
    """
    BY_NAME = "BY_NAME"
    BY_ACCESSION = "BY_ACCESSION"
    BY_LAST_UPDATE = "BY_LAST_UPDATE"
    DEFAULT = "DEFAULT"


class FilesUtil(Application):
    """
    An application to perform file management operations on Genestack.
    """
    APPLICATION_ID = 'genestack/filesUtil'

    CONTAINER = 'com.genestack.api.files.IContainerFile'
    FOLDER = 'com.genestack.api.files.IFolder'
    EXPERIMENT = 'com.genestack.bio.files.IExperiment'
    DATASET = 'com.genestack.api.files.IDataset'

    FILE = 'com.genestack.api.files.IFile'
    UNALIGNED_READS = 'com.genestack.bio.files.IUnalignedReads'
    UNALIGNED_READS_DATA = 'com.genestack.bio.files.IUnalignedReadsData'
    ALIGNED_READS = 'com.genestack.bio.files.IAlignedReads'
    VARIATION_FILE = 'com.genestack.bio.files.IVariationFile'
    APPLICATION_PAGE_FILE = 'com.genestack.api.files.IApplicationPageFile'
    REFERENCE_GENOME = 'com.genestack.bio.files.IReferenceGenome'
    AUXILIARY_FILE = 'com.genestack.api.files.IAuxiliaryFile'
    INDEX_FILE = 'com.genestack.api.files.IIndexFile'
    CODON_TABLE = 'com.genestack.bio.files.ICodonTable'
    GENOME_BED_DATA = 'com.genestack.bio.files.genomedata.IGenomeBEDData'
    GENOME_WIGGLE_DATA = 'com.genestack.bio.files.genomedata.IGenomeWiggleData'
    GENOME_ANNOTATIONS = 'com.genestack.bio.files.IGenomeAnnotations'
    HTSEQ_COUNTS = 'com.genestack.bio.files.IHTSeqCounts'
    EXTERNAL_DATABASE = 'com.genestack.bio.files.IExternalDataBase'
    PREFERENCES_FILE = 'com.genestack.api.files.IPreferencesFile'
    REPORT_FILE = 'com.genestack.api.files.IReportFile'
    RAW_FILE = 'com.genestack.api.files.IRawFile'
    MICROARRAY_ASSAY = 'com.genestack.bio.files.IMicroarrayAssay'
    MICROARRAY_DATA = 'com.genestack.bio.files.IMicroarrayData'
    SEQUENCING_ASSAY = 'com.genestack.bio.files.ISequencingAssay'
    FEATURE_LIST = 'com.genestack.bio.files.IFeatureList'
    EXPRESSION_SIGNATURE = 'com.genestack.bio.files.IGeneExpressionSignature'
    EXPRESSION_LEVELS = 'com.genestack.bio.files.IExpressionLevels'

    MAX_FILE_SEARCH_LIMIT = 2000
    MAX_RELATED_TERMS_LIMIT = 10000

    def find_reference_genome(self, organism, assembly, release):
        """
        Returns the accession of the reference genome with the specified parameters:
        ``organism``, ``assembly``, ``release``.
        If more than one or no genome is found, the corresponding exceptions are thrown.

        :param organism: organism
        :type organism: str
        :param assembly: assembly
        :type assembly: str
        :param release: release
        :type release: str
        :return: accession
        :rtype: str
        :raises: :py:class:`~odm_sdk.exceptions.GenestackServerException`
                 if more than one genome, or no genome is found
        """
        return self.invoke('findReferenceGenome', organism, assembly, release)

    def find_file_by_name(self, name, parent=None, file_class=FILE):
        """
        Finds file with specified name (ignore case!) and type.
        If no file is found ``None`` is returned.
        If more than one file is found the first one is returned.
        If the parent container is not found, the corresponding exceptions are thrown.

        :param name: file name
        :type name: str
        :param parent: parent accession, private folder is default
        :type parent: str
        :param file_class: File class to be returned, default IFile
        :type file_class: str
        :return: file accession
        :rtype: str
        """
        return self.invoke(
            'getFileByName',
            name, parent, file_class
        )

    # SA: TODO: remove this methods and change usage to findFileByName
    def find_folder_by_name(self, name, parent=None):
        return self.find_file_by_name(name, parent, self.FOLDER)

    def find_aligned_reads_file_by_name(self, name, parent=None):
        return self.find_file_by_name(name, parent, self.ALIGNED_READS)

    def find_unaligned_reads_file_by_name(self, name, parent=None):
        return self.find_file_by_name(name, parent, self.UNALIGNED_READS)

    def find_variation_file_by_name(self, name, parent=None):
        return self.find_file_by_name(name, parent, self.VARIATION_FILE)

    def find_experiment_by_name(self, name, parent=None):
        return self.find_file_by_name(name, parent, self.EXPERIMENT)

    def find_application_page_file_by_name(self, name, parent=None):
        return self.find_file_by_name(name, parent, self.APPLICATION_PAGE_FILE)

    # FIXME: use pagination in this method, see #5063
    def collect_initializable_files_in_container(self, accession):
        """
        Recursively search for all initialisable file in container.

        :param accession: accession of container
        :type accession: str
        :return: list of accessions
        :rtype: list
        """
        return self.invoke('collectInitializableFilesInContainer', accession)

    def count_file_children(self, container_accession):
        """
        Count children of a container (not recursive).
        :param container_accession: accession of container
        :type container_accession: str
        :return: number of children
        :rtype int:
        """
        return self.invoke('countFileChildren', container_accession)

    def get_file_children(self, container_accession):
        """
        Return accessions of files linked to current container.

        :param container_accession:  accession of container
        :type container_accession: str
        :return: list of accessions
        :rtype: list
        """
        all_files = []
        count = 0
        while True:
            batch = self.invoke('getFileChildren', container_accession, count, FILE_BATCH_SIZE)
            all_files += batch
            count += len(batch)
            if len(batch) < FILE_BATCH_SIZE:
                break
        return all_files

    def create_folder(self, name, parent=None, description=None, metainfo=None):
        """
        Create a folder.

        :param name: name of the folder
        :type name: str
        :param parent: if not specified, create folder in the user's private folder
        :type parent: str
        :param description: description of the folder (goes into the metainfo)
        :type description: str
        :param metainfo: additional :py:class:`~odm_sdk.Metainfo`.
            Description and accession should be specified either via arguments or in a metainfo object (but not in both).
        :type metainfo: Metainfo
        :return: accession of created folder
        """
        metainfo = metainfo or Metainfo()
        metainfo.add_string(Metainfo.NAME, name)
        if description is not None:
            metainfo.add_string(Metainfo.DESCRIPTION, description)
        return self.invoke('createFolder', parent, metainfo)

    def find_or_create_folder(self, name, parent=None):
        """
        Return the folder accession if it already exists, and create it otherwise.
        If more than one folder is found the first one is returned.

        :param name: display name
        :type name: str
        :param parent: parent accession, use home folder if None
        :type parent: str
        :return: accession of folder
        :rtype: str
        """
        return self.invoke('findOrCreateFolder', name, parent)

    def link_file(self, accession, parent):
        """
        Link a file to a folder.

        :param accession: file accession
        :type accession: str
        :param parent: parent folder accession
        :type parent: str
        :rtype: None
        """
        self.invoke('linkFiles', {accession: [parent]})

    def unlink_file(self, accession, parent):
        """
        Unlink a file from a folder.

        :param accession: file accession
        :type accession: str
        :param parent: folder accession
        :type parent: str
        :rtype: None
        """
        self.invoke('unlinkFiles', {accession: [parent]})

    def link_files(self, children_to_parents_dict):
        """
        Link files to containers.

        :param children_to_parents_dict: dictionary where keys are accessions of the files to link, and
            values are lists of accessions of the containers to link into
        :type: dict

        :rtype: None
        """
        self.invoke('linkFiles', children_to_parents_dict)

    def unlink_files(self, children_to_parents_dict):
        """
        Unlink files from containers.

        :param children_to_parents_dict: dictionary where keys are accessions of the files to unlink, and
            values are lists of accessions of the containers to unlink from
        :type children_to_parents_dict: dict[str, list[str]]
        :rtype: None
        """
        self.invoke('unlinkFiles', children_to_parents_dict)

    def add_metainfo_string_value(self, accession_list, key, value):
        """
        Add a string value to the metainfo of specified files.

        :param accession_list: list of files to be updated
        :type accession_list: list[str]
        :param key: metainfo key
        :type key: str
        :param value: string
        :type value: str
        :rtype: None
        """
        self.invoke('addMetainfoStringValue', accession_list, key, value)

    def replace_metainfo_string_value(self, accession_list, key, value):
        """
        Replace a string value in the metainfo of specified files.

        :param accession_list: list of files to be updated
        :type accession_list: list[str]
        :param key: metainfo key
        :type key: str
        :param value: string
        :type value: str
        :rtype: None
        """
        self.invoke('replaceMetainfoStringValue', accession_list, key, value)

    def replace_metainfo_value(self, accession_list, key, value):
        """
        Replace a value in the metainfo of specified files.

        :param accession_list: list of files to be updated
        :type accession_list: list[str]
        :param key: metainfo key
        :type key: str
        :param value: metainfo value
        :type value: MetainfoScalarValue
        :rtype: None
        """
        self.invoke('replaceMetainfoValue', accession_list, key, value)

    def remove_metainfo_value(self, accession_list, key):
        """
        Delete a key from the metainfo of specified files.

        :param accession_list: list of files to be updated
        :type accession_list: list[str]
        :param key: metainfo key
        :type key: str
        :rtype: None
        """
        self.invoke('removeMetainfoValue', accession_list, key)

    def add_metainfo_values(self, accession, metainfo, skip_existing_keys=True, replace_existing_keys=False):
        """
        Add metainfo to a specified file.
        By default, metainfo keys that are already present in the file will be skipped.

        :param accession: accession of the file to update
        :param metainfo: metainfo object containing the metainfo to add
        :type metainfo: Metainfo
        :param skip_existing_keys: ignore metainfo keys that are already present in the file's metainfo
            (default: ``True``)
        :type skip_existing_keys: bool
        :param replace_existing_keys: replace the existing metainfo value for the metainfo keys
            that are already present in the file's metainfo (default: ``False``)
        :type replace_existing_keys: bool
        :rtype: None
        """
        self.invoke('addMetainfoValues', accession, metainfo, skip_existing_keys, replace_existing_keys)

    def get_metainfo_values_as_strings(self, accessions_list, keys_list=None):
        """
        Retrieve metainfo values as strings for specific files and metainfo keys.
        Metainfo value lists are concatenated to string using ', ' as delimiter.
        The function returns a dictionary.

        :param accessions_list: accessions of the files to retrieve
        :type: accessions: list[str]
        :param keys_list: metainfo keys to retrieve (if ``None``, all non-technical keys are retrieved for each file)
        :type: keys: list[str]|None
        :return: a two-level dictionary with the following structure: accession -> key -> value
        :rtype: dict[str, dict[str, str]]
        """
        return self.invoke('getMetainfoValuesAsStrings', accessions_list, keys_list)

    def get_metainfo_values_as_string_list(self, accessions_list, keys_list=None):
        """
        Retrieve metainfo values as lists of strings for specific files and metainfo keys.
        The function returns a dictionary.

        :param accessions_list: accessions of the files to retrieve
        :type: accessions: list[str]
        :param keys_list: metainfo keys to retrieve (if ``None``, all non-technical keys are retrieved for each file)
        :type: keys: list[str]|None
        :return: a two-level dictionary with the following structure: accession -> key -> value list
        :rtype: dict[str, dict[str, list[str]]]
        """
        return self.invoke('getMetainfoValuesAsStringList', accessions_list, keys_list)

    def get_special_folder(self, name):
        """
        Return the accession of a special folder.

        Available special folders are described in :class:`.SpecialFolders`

        :param name: special folder name
        :type name: str
        :return: accession
        :rtype: str
        :raises: GenestackException: if folder name is unknown
        """
        special_folders = (SpecialFolders.IMPORTED, SpecialFolders.CREATED, SpecialFolders.TEMPORARY,
                           SpecialFolders.UPLOADED, SpecialFolders.MY_DATASETS)
        if name not in special_folders:
            raise GenestackException("Name '%s' must be one of %s" % (name, ', '.join(special_folders)))
        return self.invoke('getSpecialFolder', name)

    def get_group_folder_info(self, group_accession):
        raise NotImplementedError("FilesUtil.get_group_folder_info has been removed in v0.33")

    def get_folder(self, parent, *names, **kwargs):
        """
        Find a subfolder (by name) in a folder passed as an accession,
        returning accession of that subfolder.  If several names are provided,
        treat them as a path components for the sub-sub-...-folder down the
        folder hierarchy, returning accession of that deepmost folder:

        - ``fu.get_folder('GS777', 'RNASeq')`` looks for subfolder with *name*
          "RNASeq" in folder with *accession* "GS777", and returns accession of
          that "RNASeq" subfolder;

        - ``fu.get_folder('GS777', 'Experiments', 'RNASeq')`` looks for
          subfolder with *name* "Experiments" in a folder with *accession*
          "GS777", then looks for "RNASeq" in "Experiments", and returns the
          accession of "RNASeq".

        If ``create=True`` is passed as a *kwarg*, all the folders in ``names``
        hierarchy will be created (otherwise ``GenestackException`` is raised).

        :param parent: accession of folder to search in
        :type parent: str
        :param \*names: tuple of "path components", a hierarchy of folders to
                        find
        :type names: tuple
        :param create: whether to create folders from ``names`` if they don't
                       exist or not; default is ``False`` (raise
                       ``GenestackException`` if any folder doesn't exist)
        :type create: bool
        :return: accession of found (or created) subfolder
        :rtype: str
        :raises GenestackException:
            if no name is passed, or folder with required name is not found
            (and shouldn't be created)
        """
        if not names:
            raise GenestackException("At least one path should be specified")

        create = bool(kwargs.get('create'))
        for path in names:
            if create:
                parent = self.find_or_create_folder(path, parent=parent)
            else:
                _parent_accession = self.find_file_by_name(path, parent=parent, file_class=self.FOLDER)
                if _parent_accession is None:
                    raise GenestackException('Cannot find folder with name "%s" '
                                             'in folder with accession: %s'
                                             % (path, parent))
                parent = _parent_accession
        return parent

    def get_home_folder(self):
        """
        Return the accession of the current user's home folder.

        :return: accession of home folder
        :rtype: str
        """
        return 'private'

    def get_public_folder(self):
        """
        Return the accession of the ``Public`` folder on the current Genestack instance.

        :return: accession of ``Public`` folder
        :rtype: str
        """
        return 'public'

    def get_infos(self, accession_list):
        """
        Returns a list of dictionaries with information about each of the specified files.
        This will return an error if any of the accessions is not valid.
        The order of the returned list is the same as the one of the accessions list.

        The information dictionaries have the following structure:

            - accession
            - owner
            - name
            - isDataset
            - application

                - id

            - initializationStatus

                - isError
                - id

            - permissionsByGroup (the value for each key is a dictionary with group accessions as keys)

                - groupNames
                - ids

            - time

                - fileCreation
                - initializationQueued
                - initializationStart
                - initializationEnd
                - fileCreation
                - lastMetainfoModification


        :param accession_list: list of valid accessions.
        :type accession_list: list
        :return: list of file info dictionaries.
        :rtype: list[dict[str, object]]
        """
        return self.invoke('getInfos', accession_list)

    def rename_file(self, accession, name):
        """
        Rename a file.

        :param accession: file accession
        :type accession: str
        :param name: name
        :type name: str
        :rtype: None
        """
        self.invoke('renameFile', accession, name)

    def mark_for_tests(self, app_file):
        """
        Mark Genestack file as test one by adding corresponding key to metainfo.
        Test file will calculate md5 checksums of its encapsulated physical files
        during initialization.

        :param app_file: accession of file
        :return: None
        """
        metainfo = Metainfo()
        metainfo.add_boolean(CALCULATE_CHECKSUMS_KEY, True)
        self.add_metainfo_values(app_file, metainfo)

    def mark_obsolete(self, accession):
        """
        Mark Genestack file as obsolete one by adding corresponding key to metainfo.

        :param accession: accession of file
        :return: None
        """
        metainfo = Metainfo()
        metainfo.add_boolean('genestack:obsolete', True)
        self.add_metainfo_values(accession, metainfo)

    def add_checksums(self, app_file, expected_checksums):
        """
        Add expected MD5 checksum to the metainfo of a CLA file.
        Expected checksums are calculated in the following way:

            - The number of checksums equals number of entries in storage.
              For instance, a Reference Genome file has 2 entries (annotation and sequence files).
            - If there are multiple files in one entry, they will be concatenated in the same order
              as they were ``PUT`` to storage by the initialization script.
            - If a file is marked for testing, then after initialization its metainfo
              will contain both expected and actual checksum values.

        :param app_file: accession of application file
        :param expected_checksums: collection of MD5 checksums
        :return: None
        """
        metainfo = Metainfo()
        for key, value in expected_checksums.items():
            metainfo.add_string('%s%s' % (EXPECTED_CHECKSUM_PREFIX, key), value)
        self.add_metainfo_values(app_file, metainfo)

    def find_files(
            self,
            file_filter,
            sort_order=SortOrder.DEFAULT,
            ascending=False,
            offset=0,
            limit=MAX_FILE_SEARCH_LIMIT
    ):
        """
        Search for files with ``file_filter`` and return dictionary with two key/value pairs:

         - ``'total'``: total number (``int``) of files matching the query
         - ``'result'``: list of file info dictionaries for subset of matching files
                         (from ``offset`` to ``offset+limit``). See the documentation of
                         :py:meth:`~odm_sdk.FilesUtil.get_infos` for the structure
                         of these objects.

        :param file_filter: file filter
        :type file_filter: FileFilter
        :param sort_order: sorting order for the results,
                           see :py:class:`~odm_sdk.files_util.SortOrder`
        :type sort_order: str
        :param ascending: should the results be in ascending order? (default: False)
        :type ascending: bool
        :param offset: search offset (default: 0, cannot be negative)
        :type offset: int
        :param limit: maximum number of results to return (max and default: 100)
        :type limit: int
        :return: a dictionary with search response
        :rtype: dict[str, int|list[dict[str, str|dict]]]
        """
        limit = min(self.MAX_FILE_SEARCH_LIMIT, limit)
        if offset < 0 or limit < 0:
            raise GenestackException("Search offset/limit cannot be negative")
        if not validate_constant(SortOrder, sort_order):
            raise GenestackException("Invalid sort order")
        return self.invoke('findFiles', file_filter.get_dict(), sort_order, ascending, offset, limit)

    def collect_metainfos(self, accessions):
        """
        Get complete metainfo of a list of files.

        :param accessions: list of accessions
        :type accessions: list[str]
        :return: list of metainfo objects
        :rtype: list[Metainfo]
        """
        return [Metainfo.parse_metainfo_from_dict(mi)
                for mi in self.invoke('getMetainfo', accessions)]
    def initialize(self, accessions):
        """
        Start initialization for the specified accessions.
        Missed accession and initialization failures are ignored silently.

        :param list[str] accessions: list of accessions
        :rtype: None
        """
        self.invoke('initialize', accessions)

    # TODO deprecate and use FilesUtil(self.connection).get_infos(accessions) instead
    def load_info(self, accessions):
        """
        Takes as input a list of file accessions and returns a list of dictionaries (one for each accession)
        with the following structure:

            - accession: (str) file accession
            - name: (str) file name if the file exists
            - status: (str) initialization status

        The possible values for ``status`` are:

            - NoSuchFile
            - NotApplicable
            - NotStarted
            - InProgress
            - Complete
            - Failed


        :param list[str] accessions: list of accessions
        :return: list of dictionaries
        :rtype: list
        """
        return self.invoke('loadInfo', accessions)

    def search_files(self, accession):
        return self.invoke('searchFiles',
                           accession,  # type: str
                           {}, 100, 0, None, False)



# -*- coding: utf-8 -*-

#
# Copyright (c) 2011-2023 Genestack Limited
# All Rights Reserved
# THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF GENESTACK LIMITED
# The copyright notice above does not evidence any
# actual or intended publication of such source code.
#

import itertools
import os
from odm_sdk import make_connection_parser

PLACEHOLDER_VAR = 'x'

# Metric prefixes
DIV_PREFIXES = [
    ('femto', 'f', -15),
    ('pico', 'p', -12),
    ('nano', 'n', -9),
    ('micro', 'μ', -6),
    ('milli', 'm', -3),
]
MUL_PREFIXES = [
    ('kilo', 'k', 3),
    ('mega', 'M', 6),
    # ('giga', 'G', 9),
    # ('tera', 'T', 12),
    # ('peta', 'P', 15),
]
NON_C_PREFIXES = [
    ('centi', 'c', -2),
]


class Unit(object):
    """
    Base class for the unit hierarchy. Represents a generic unit.
    """

    def __init__(self, label, name, parent=None, factor=1, conversion_to=None,
                 conversion_from=None):
        """
        :param label: unit label. Used to build relations between units.
        :type label: str
        :param name: unit abbreviation.
        :type name: str
        :param parent: parent unit. Used to convert units between each other.
        :type parent: Unit
        :param factor: a number that can be used to convert current unit to parent unit by multiplying
        :type factor: int | long | float
        :param conversion_to: formula to convert this unit to it's parent unit
        :type conversion_to: str
        :param conversion_from: formula to convert from parent unit to this unit
        :type conversion_from: str
        """

        self.label = label
        self.name = name
        self.parent = parent
        self.factor = factor
        self._conversion_to = conversion_to
        self._conversion_from = conversion_from

    def to_csv(self):
        return (
            self.label,
            self.name,
            '' if self.parent is None else self.parent.label,
            self.conversion_to,
            self.conversion_from
        )

    @property
    def str_factor(self):
        return str(self.factor)

    @property
    def conversion_to(self):
        if self.parent is None:
            return ''
        if self._conversion_to is None:
            f = self.str_factor
            return '%s * %s' % (PLACEHOLDER_VAR, f) if f else ''
        else:
            return self._conversion_to

    @property
    def conversion_from(self):
        if self.parent is None:
            return ''
        if self._conversion_from is None:
            f = self.str_factor
            return '%s / %s' % (PLACEHOLDER_VAR, f) if f else ''
        else:
            return self._conversion_from


class SimpleUnit(Unit):
    """
    Represents units that can be converted between each other by simply multiplying with a power of 10.
    """

    def __init__(self, label, name, parent=None, power_of_10=0):
        super(SimpleUnit, self).__init__(label, name, parent)
        self.power_of_10 = power_of_10

    def apply_prefix(self, prefix, name, power_of_10):
        return SimpleUnit(
            label=prefix + self.label,
            name=name + self.name,
            parent=self if self.parent is None else self.parent,
            power_of_10=self.power_of_10 + power_of_10
        )

    @property
    def str_factor(self):
        return '1' if self.power_of_10 == 0 else '10 ^ %d' % self.power_of_10


class PowerUnit(Unit):
    """
    Represents units that are actually a power of another unit (like square metres).
    """

    def __init__(self, base, power, parent=None):
        def power_to_str():
            if power == 2:
                return 'square'
            elif power == 3:
                return 'cubic'

        super(PowerUnit, self).__init__(
            label='%s %s' % (power_to_str(), base.label),
            name='%s^%d' % (base.name, power),
            parent=parent,
        )
        self.base = base
        self.power = power

    def apply_prefix(self, *args):
        return PowerUnit(
            self.base.apply_prefix(*args),
            self.power,
            self if self.parent is None else self.parent
        )

    @property
    def str_factor(self):
        if self.parent is None:
            return '1'
        if isinstance(self.base, SimpleUnit):
            return '10 ^ %d' % (self.base.power_of_10 * self.power)
        return '(%s) ^ %d' % (super(PowerUnit, self).str_factor, self.power)


class CombinationUnit(Unit):
    """
    Represents units that are expressed as a relation between several units (like metres per second).
    Only division relation is supported now.
    """

    def __init__(self, *args):
        def compute_parent():
            if all([arg.parent is None for arg in args]):
                return None
            return CombinationUnit(*[arg if arg.parent is None else arg.parent for arg in args])

        super(CombinationUnit, self).__init__(
            label=' per '.join(a.label for a in args),
            name='/'.join(a.name if a.name else '1' for a in args),
            parent=compute_parent()
        )
        self.units = args

    @property
    def str_factor(self):
        def get_format_operands():
            return ' / '.join(unit.str_factor for unit in self.units)

        return '' if self.parent is None else '(%s)' % get_format_operands()


class DictionaryGenerator(object):
    HEADERS = ['Label', 'Name', 'Parent', 'To_parent', 'From_parent']

    def __init__(self, dict_name):
        super(DictionaryGenerator, self).__init__()
        self._dict_name = dict_name

    def __enter__(self):
        self._dict_file = open(self._dict_name, 'w')
        self._dict_file.write(','.join(self.HEADERS) + '\n')
        return self

    # noinspection PyUnusedLocal
    def __exit__(self, *args):
        self._dict_file.close()

    def save_unit_array(self, arr):
        self._dict_file.writelines(map(lambda u: ','.join(u.to_csv()) + '\n', arr))

    def save_unit(self, unit):
        self.save_unit_array([unit])


def generate_unit_array(unit, prefixes):
    result = [unit]
    for prefix in prefixes:
        result.append(unit.apply_prefix(*prefix))
    return result


# Predefined units
METRE = SimpleUnit('metre', 'm')
SQ_METRE = PowerUnit(METRE, 2)
CUBIC_METRE = PowerUnit(METRE, 3)
LITRE = SimpleUnit('litre', 'l', parent=CUBIC_METRE, power_of_10=-3)
SECOND = Unit('second', 'sec')
METRES = generate_unit_array(METRE, DIV_PREFIXES + MUL_PREFIXES[:-1] + NON_C_PREFIXES)
SQUARE_METRES = generate_unit_array(SQ_METRE, DIV_PREFIXES + NON_C_PREFIXES)
VOLUMES = generate_unit_array(CUBIC_METRE, DIV_PREFIXES + NON_C_PREFIXES) + generate_unit_array(
    LITRE, DIV_PREFIXES + NON_C_PREFIXES)
TIME_UNITS = [
    SECOND,
    Unit('week', 'w', SECOND, 7 * 24 * 60 * 60),
    Unit('day', 'd', SECOND, 24 * 60 * 60),
    Unit('hour', 'h', SECOND, 60 * 60),
    Unit('minute', 'min', SECOND, 60)
]


def get_combinations(*args):
    return (CombinationUnit(*combination) for combination in itertools.product(*args))


def generate_dose_dict(dict_name):
    gram = SimpleUnit('gram', 'g')
    mole = SimpleUnit('mole', 'mol')
    molar = SimpleUnit('molar', 'M')

    weights = generate_unit_array(gram, DIV_PREFIXES + MUL_PREFIXES[:-1])
    moles = generate_unit_array(mole, DIV_PREFIXES)
    molars = generate_unit_array(molar, DIV_PREFIXES)
    with DictionaryGenerator(dict_name) as dictionary:
        dictionary.save_unit_array(weights)
        dictionary.save_unit_array(METRES)
        dictionary.save_unit_array(SQUARE_METRES)
        dictionary.save_unit_array(VOLUMES)
        dictionary.save_unit_array(TIME_UNITS)
        dictionary.save_unit_array(moles)
        dictionary.save_unit_array(molars)

        dictionary.save_unit_array(get_combinations(weights, weights))
        dictionary.save_unit_array(get_combinations(weights, VOLUMES))
        dictionary.save_unit_array(get_combinations(VOLUMES, weights))
        dictionary.save_unit_array(get_combinations(VOLUMES, VOLUMES))
        dictionary.save_unit_array(get_combinations(weights, TIME_UNITS))
        dictionary.save_unit_array(get_combinations(weights, SQUARE_METRES))
        dictionary.save_unit_array(get_combinations(weights, weights, TIME_UNITS))
        dictionary.save_unit_array(get_combinations(moles, weights))


def generate_energy_dict(dict_name):
    sievert = SimpleUnit('sievert', 'Sv')
    watt = SimpleUnit('watt', 'W')
    gray = SimpleUnit('gray', 'Gy')
    kelvin = Unit('kelvin', 'K')
    joule = SimpleUnit('joule', 'J')
    becquerel = SimpleUnit('becquerel', 'Bq')
    rad = SimpleUnit('rad', 'rad')
    einstein = SimpleUnit('einstein', 'E')
    lux = SimpleUnit('lux', 'lx')
    pascal = SimpleUnit('pascal', 'Pa')

    watts = generate_unit_array(watt, DIV_PREFIXES + MUL_PREFIXES)
    grays = generate_unit_array(gray, DIV_PREFIXES)
    temperatures = (
        kelvin,
        Unit(
            'degree Celsius', '°C', kelvin,
            conversion_to='%s + 273.15' % PLACEHOLDER_VAR,
            conversion_from='%s - 273.15' % PLACEHOLDER_VAR
        ),
        Unit(
            'degree Fahrenheit', '°F', kelvin,
            conversion_to='(%s + 459.67) * 5 / 9' % PLACEHOLDER_VAR,
            conversion_from='%s * 9 / 5 - 459.67' % PLACEHOLDER_VAR
        )
    )
    joules = generate_unit_array(joule, DIV_PREFIXES)
    becquerels = generate_unit_array(becquerel, MUL_PREFIXES)
    einsteins = generate_unit_array(einstein, MUL_PREFIXES)
    with DictionaryGenerator(dict_name) as dictionary:
        dictionary.save_unit_array(watts)
        dictionary.save_unit_array(grays)
        dictionary.save_unit_array(temperatures)
        dictionary.save_unit_array(joules)
        dictionary.save_unit_array(becquerels)

        dictionary.save_unit_array(generate_unit_array(sievert, DIV_PREFIXES))
        dictionary.save_unit_array(generate_unit_array(rad, DIV_PREFIXES))
        dictionary.save_unit_array(generate_unit_array(lux, DIV_PREFIXES + MUL_PREFIXES))
        dictionary.save_unit_array(generate_unit_array(pascal, DIV_PREFIXES + MUL_PREFIXES))

        dictionary.save_unit_array(get_combinations(watts, SQUARE_METRES))
        dictionary.save_unit_array(get_combinations(grays, TIME_UNITS))
        dictionary.save_unit_array(get_combinations(joules, SQUARE_METRES))
        dictionary.save_unit_array(get_combinations(becquerels, VOLUMES))
        dictionary.save_unit_array(get_combinations(einsteins, TIME_UNITS, SQUARE_METRES))


def generate_length_dict(dict_name):
    with DictionaryGenerator(dict_name) as dictionary:
        dictionary.save_unit_array(METRES)
        dictionary.save_unit(Unit('inch', 'in', METRE, 0.0254))


def generate_time_dict(dict_name):
    month = Unit('month', 'm', SECOND, 24 * 60 * 60 * 365 / 12)
    year = Unit('year', 'y', SECOND, 24 * 60 * 60 * 365)
    with DictionaryGenerator(dict_name) as dictionary:
        dictionary.save_unit_array(TIME_UNITS)
        dictionary.save_unit(month)
        dictionary.save_unit(year)
        dictionary.save_unit(Unit('hertz', 'Hz'))

        dictionary.save_unit(CombinationUnit(TIME_UNITS[0], TIME_UNITS[0]))
        dictionary.save_unit(CombinationUnit(TIME_UNITS[3], TIME_UNITS[2]))


def generate_bio_counts_dict(dict_name):
    colony_forming_unit = Unit('colony_forming_unit', '')
    cells = Unit('cells', '')

    with DictionaryGenerator(dict_name) as dictionary:
        dictionary.save_unit(colony_forming_unit)

        dictionary.save_unit_array(get_combinations([colony_forming_unit], TIME_UNITS))
        dictionary.save_unit_array(get_combinations([colony_forming_unit], VOLUMES))
        dictionary.save_unit_array(get_combinations([cells], VOLUMES))


def main():

    args = get_args()

    if args.destination_path:
        destination_path = args.destination_path
    else:
        destination_path = os.path.join(os.getcwd(), 'generated')

    if not os.path.exists(destination_path):
        os.mkdir(destination_path)

    def make(generator, basename):
        """Make dictionary and return path of created file."""
        path = os.path.join(destination_path, basename)
        generator(path)
        return path

    make(generate_dose_dict, 'units_dose.csv')
    make(generate_energy_dict, 'units_energy.csv')
    make(generate_length_dict, 'units_length.csv')
    make(generate_time_dict, 'units_time.csv')
    make(generate_bio_counts_dict, 'units_bio_counts.csv')


def get_args():
    parser = make_connection_parser()
    parser.add_argument('--destination-path',
                        help="[Optional] Path to put generated dictionaries, "
                        "The path will be generated automatically when the parameter is not specified. ")

    return parser.parse_args()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

#
# Copyright (c) 2011-2020 Genestack Limited
# All Rights Reserved
# THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF GENESTACK LIMITED
# The copyright notice above does not evidence any
# actual or intended publication of such source code.
from __future__ import print_function, unicode_literals

import csv
import sys

PY3 = sys.version_info > (3,)

BLUE = 34
GREEN = 32
RED = 31


def colored(text, color):
    if sys.platform == 'win32' or not sys.stdout.isatty():
        return text
    return '\033[%dm%s\033[0m' % (color, text)


class TsvReader():
    def __init__(self, filename):
        self.filename = filename
        self.dialect = csv.excel_tab
        self.encoding = 'utf-8'
        self.f = None
        self.reader = None

    def __enter__(self):
        if PY3:
            self.f = open(self.filename, 'rt',
                          encoding=self.encoding, newline='')
        else:
            self.f = open(self.filename, 'rb')
        self.reader = csv.reader(self.f, dialect=self.dialect)
        return self

    def __exit__(self, type_, value, traceback):
        self.f.close()

    def next(self):
        row = next(self.reader)
        if PY3:
            return row
        return [s.decode(self.encoding) for s in row]

    __next__ = next

    def __iter__(self):
        return self



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# This script adds permissions to existing users.
# It requires tsv file groups_and_users.tsv to be present in the same directory as this script.
# groups_and_users.tsv should contain lines of the next format: groups, separated by commas,
# and users, separated by commas.
# Groups and users should be separated by tabs.

# Same groups may be repeated several times on different lines.

# For example:
# MANAGE_ORGANIZATION,MANAGE_GROUPS,MANAGE_TEMPLATES,MANAGE_FACETS alice@alphacorp.com,suzy@alphacorp.com
# MANAGE_ORGANIZATION  bob@alphacorp.com

# See the full instruction at:
# https://genestack.atlassian.net/wiki/spaces/ODMP/pages/1143242753/How+to+create+groups+and+add+existing+users+to+them+via+python+scripts
from __future__ import print_function, unicode_literals


import sys

from odm_sdk.utils import make_connection_parser, get_connection
from odm_sdk.scripts.utils import colored, TsvReader, GREEN, RED


def get_permissions_from_odm(connection):
    return connection.application('genestack/usersadmin-api').invoke('getAvailableDomainPermissions')


def add_permissions(connection, email, permissions):
    return connection.application('genestack/usersadmin-api').invoke(
        'setUserDomainPermissions', email, permissions)


def read_data_for_updating_users(args):
    users_for_updating = []
    incorrect_rows = []
    rows_number = 0
    with TsvReader(args.file_with_users) as usersfile:
        for row in usersfile:
            rows_number += 1
            try:
                permissions, emails = row[0].strip(), row[1].strip()
                if len(permissions) == 0 or len(emails) == 0:
                    raise IndexError
            except IndexError:
                incorrect_rows.append(row)
                continue
            user_permissions = [permission.strip() for permission in permissions.split(",") if permission]

            users_for_updating += [
                {'email': email.strip(), 'permissions': user_permissions} for email in emails.split(",")
                if email
            ]

        incorrect_rows_number = len(incorrect_rows)
        if incorrect_rows_number > 0:
            print(colored("Error: each line should contain groups separated by comma and users' "
                          "emails separated by comma, groups and users should be separated by tab.",
                          RED))
            if rows_number != incorrect_rows_number:
                print(colored("Lines to correct:", RED))
                for incorrect_row in incorrect_rows:
                    print(colored(incorrect_row, RED))
            print(colored("FAILED: Correct data and rerun this script.", RED))
            sys.exit()
    return users_for_updating


def check_all_users_exist(connection, all_users_to_be_added):
    existing_users = connection.application('genestack/usersadmin-api').invoke('listUsers')
    existing_emails = {existing_user['email'] for existing_user in existing_users}
    non_existing_emails = {
        user['email'] for user in all_users_to_be_added
        if user['email'] not in existing_emails
    }
    return non_existing_emails


def check_all_permissions_id_exists_in_odm(connection, users_for_updating):
    permission_ids_in_odm = []
    for i in get_permissions_from_odm(connection):
        permission_ids_in_odm.append(i["id"])
    permission_ids_from_file = []
    for i in users_for_updating:
        for j in i["permissions"]:
            permission_ids_from_file.append(j)
    non_existing_permissions = []
    for i in permission_ids_from_file:
        if i not in permission_ids_in_odm:
            non_existing_permissions.append(i)
    return non_existing_permissions


def main():
    parser = make_connection_parser()
    parser.add_argument('--file_with_users', metavar='<file_with_users>',
                        default="permissions_and_users.tsv",
                        help='add permissions to existing users')
    connection = get_connection()
    args = parser.parse_args()
    users_for_updating = read_data_for_updating_users(args)

    non_existing_emails = check_all_users_exist(connection, users_for_updating)
    all_users_exist = len(non_existing_emails) == 0

    non_existing_permissions = check_all_permissions_id_exists_in_odm(connection, users_for_updating)
    all_permissions_exist = len(non_existing_permissions) == 0

    for user in users_for_updating:
        add_permissions(connection, user["email"], user["permissions"])

    if all_permissions_exist and all_users_exist:
        print(colored("Success", GREEN))
    else:
        if not all_permissions_exist:
            print(colored("Permissions %s don't exist" % non_existing_permissions, RED))
        if not all_users_exist:
            print(colored("Users %s don't exist" % non_existing_emails, RED))
        if not all_users_exist or not all_permissions_exist:
            print(colored("Create missed items and rerun this script with the same input file",
                          RED))


if __name__ == "__main__":
    main()



class Permissions(object):

    _JAVA_PKG = "com.genestack.file."

    FILE_ACCESS = _JAVA_PKG + "access"
    FILE_READ_CONTENT = _JAVA_PKG + "readContent"
    FILE_WRITE = _JAVA_PKG + "write"
    FILE_CLONE_DATA = _JAVA_PKG + "cloneData"



#  Copyright (c) 2011-2024 Genestack Limited
#  All Rights Reserved
#  THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF GENESTACK LIMITED
#  The copyright notice above does not evidence any
#  actual or intended publication of such source code.

import contextlib
import re
import unittest
from argparse import Namespace
from io import StringIO

import pytest

from odm_sdk import get_user, make_connection_parser


class ArgumentsTest(unittest.TestCase):

    def test_parse_default(self):
        parser = make_connection_parser()
        args = parser.parse_args()
        expected = Namespace(debug=False, host=None, pwd=None, show_logs=False,
                             user=None, token=None, access_token=None)
        self.assertEqual(expected, args)

    def test_password_without_user(self):
        parser = make_connection_parser()
        error_message = StringIO()
        with pytest.raises(SystemExit) as se, contextlib.redirect_stderr(error_message):
            parser.parse_args(['-p', 'some_password'])

        # Test stderr output that was written by parser before raising error:

        # Expected output consist of usage section and line with filename and tested error
        expected_error_message = 'Password should not be specified without user'

        # Save parser.print_usage output to the variable
        f = StringIO()
        parser.print_usage(file=f)

        expected_output = f.getvalue()
        test_runner = re.search(r'^usage: (\S+)?', f.getvalue()).group(1)
        expected_output += '%s: error: %s\n' % (test_runner, expected_error_message)

        self.assertEqual(SystemExit, se.type)
        self.assertEqual(2, se.value.code)
        self.assertEqual(expected_output, error_message.getvalue())

    def test_token_and_user(self):
        parser = make_connection_parser()

        error_message = StringIO()
        with pytest.raises(SystemExit) as se, contextlib.redirect_stderr(error_message):
            parser.parse_args(['-u', 'some_user', '--token', 'some_token'])

        # Test stderr output that was written by parser before raising error:

        # Expected output consist of usage section and line with filename and tested error
        expected_error_message = 'Exactly one of token, access_token or user should be specified'

        # Save parser.print_usage output to the variable
        f = StringIO()
        parser.print_usage(file=f)

        expected_output = f.getvalue()
        test_runner = re.search(r'^usage: (\S+)?', f.getvalue()).group(1)
        expected_output += '%s: error: %s\n' % (test_runner, expected_error_message)

        self.assertEqual(SystemExit, se.type)
        self.assertEqual(2, se.value.code)
        self.assertEqual(expected_output, error_message.getvalue())

    def test_token_without_username(self):
        parser = make_connection_parser()
        some_token = 'some_token'
        args = parser.parse_args(['--token', some_token])
        user = get_user(args)
        self.assertEqual(some_token, user.token)


if __name__ == '__main__':
    unittest.main()



from odm_sdk import Application, GenestackException


class _BaseExpressionNavigator(Application):
    def _create_file(self, groups, organism=None, normalized_input=None, options=None):
        assignments = [(group_id, accession) for group_id, group in enumerate(groups, 1)
                       for accession in group['accessions']]
        organism = organism or "unknown organism"

        group_names = [group.get("name") or "Group %d" % i for i, group in enumerate(groups, 1)]
        group_names = ["Ungrouped"] + group_names
        group_descriptions = [group.get("description") or "No description provided" for group in groups]
        group_descriptions = ["Ungrouped"] + group_descriptions

        options = options or {}
        params = {
            'accessionList': [acc for group_id, acc in assignments],
            'groupIdList': [str(group_id) for group_id, acc in assignments],
            'organism': organism,
            'groupsNameList': group_names,
            'groupsDescriptionList': group_descriptions,
            'programOptions': options,
        }

        if normalized_input:
            params['sourcesAccessions'] = normalized_input
        return self.invoke('createFile', params)

    def get_differential_expression_stats(self, accessions_to_queries):
        """
        Get differential expression statistics from files.
        This method returns a dictionary where values are list of statistics dictionaries.
        Each statistics dictionary has the following structure: (all fields are always present)

            - pValue
            - logFoldChange
            - adjustedPValue
            - genomeFeature (dict):

                - featureType
                - featureId
                - subregions (list)  (``[]`` if unavailable)
                - featureName
                - location (dict):

                    - to (``-1`` if unavailable)
                    - from (0-based, ``-1`` if unavailable)
                    - contigName (``''`` if unavailable)
                    - normalizedContigName (``''`` if unavailable)

                - parentId  (``None`` if unavailable)
                - attributes (dict) (``{}`` if unavailable)

            - contrastLevel

        :param accessions_to_queries: a dictionary whose keys are accessions of differential expression files,
                                      and whose values are ``GenomeQuery`` objects
        :type accessions_to_queries: dict[GenomeQuery]
        :return: dictionary whose keys are file accessions and values are lists of dictionaries with the format
                 described above
        :rtype: dict[list[dict]]
        """
        return self.invoke('getDifferentialExpressionStats', {acc: query.get_map() for acc, query in
                                                              accessions_to_queries.items()})


class ExpressionNavigatorforMicroarrays(_BaseExpressionNavigator):
    APPLICATION_ID = 'genestack/expressionNavigator-microarrays'

    def create_file(self, groups, normalized_microarray_file, microarray_annotation, organism=None):
        """
        Create an Expression Navigator file from a normalized microarray file.
        Each group is described by a dictionary with the following keys:

            - ``accessions``: list of accessions of the source microarray files for this group
            - ``name`` (optional): group name
            - ``description`` (optional): group description
            - ``is_control`` (optional): boolean value indicating whether the group is a control group

        :param groups: list of dictionaries describing the groups for differential expression.
                       See above for the dictionary structure.
        :param normalized_microarray_file: accession of normalized microarray file
        :param microarray_annotation: accession of the microarray annotation file
        :param organism: organism
        :return: accession of the created Expression Navigator file
        """
        control_group = None
        for i, group in enumerate(groups, 1):
            if group.get("is_control"):
                control_group = str(i)
                break
        options = {
            'controlGroupOption': {'value': control_group or "None"},
            'microarrayAnnotationSource': {'type': 'source', 'value': microarray_annotation}
        }
        return self._create_file(groups, organism=organism, normalized_input=[normalized_microarray_file],
                                 options=options)


class ExpressionNavigatorforGenes(_BaseExpressionNavigator):
    APPLICATION_ID = 'genestack/expressionNavigator'

    PKG_EDGER = "edgeR"
    PKG_DESEQ = "DESeq2"

    def create_file(self, groups, r_package=PKG_DESEQ, organism=None):
        """
        Create an expression navigator file from RNA-seq gene counts files.
        Each group is described by a dictionary with the following keys:

            - ``accessions``: list of accessions of the raw gene counts files for this group
            - ``name`` (optional): group name
            - ``description`` (optional): group description

        :param groups: list of dictionaries describing the groups for differential expression.
                       See above for the dictionary structure.
        :param r_package: name of R package to use for differential expression (either ``edgeR`` or ``DESeq2``)
        :param organism: organism
        :return: accession of the created Expression Navigator file
        """
        if r_package not in {self.PKG_DESEQ, self.PKG_EDGER}:
            raise GenestackException("Invalid package option for differential expression '%s'" % r_package)
        options = {'package': {'value': r_package}}
        return self._create_file(groups, organism=organism, options=options)


class ExpressionNavigatorforIsoforms(_BaseExpressionNavigator):
    APPLICATION_ID = 'genestack/expressionNavigator-isoforms'

    def create_file(self, groups, fragment_bias_corr=True, multi_mapping_corr=True, organism=None):
        """
        Create an expression navigator file from RNA-seq isoform FPKM counts files.
        Each group is described by a dictionary with the following keys:

            - ``accessions``: list of accessions of the isoform counts files for this group
            - ``name`` (optional): group name
            - ``description`` (optional): group description

        :param groups: list of dictionaries describing the groups for differential expression.
                       See above for the dictionary structure.
        :param fragment_bias_corr: apply correction for fragment bias
        :type fragment_bias_corr: bool
        :param multi_mapping_corr: apply correction for reads with multiple mappings
        :type multi_mapping_corr: bool
        :param organism: organism
        :return: accession of the created Expression Navigator file
        """
        options = {
            'fragmentBiasCorrectOption': {'value': 'Yes' if fragment_bias_corr else 'No'},
            'multiReadsCorrectOption': {'value': 'Yes' if multi_mapping_corr else 'No'}
        }
        return self._create_file(groups, organism=organism, options=options)
